#!/usr/bin/python3

"""
Code modified from https://github.com/AayushXtreme/gitdir which is itself \
a modification of https://github.com/sdushantha/gitdir

MIT License

Copyright (c) 2019 Siddharth Dushantha

"""

import re
import os
import urllib.request
import argparse
import json
import sys
import random
from pathlib import Path


def get_proxy(proxies=None, _print=False):
    proxy = urllib.request.ProxyHandler({})
    if proxies is not None:
        option = 'http://' + random.choice(proxies)
        if _print:
            print(f'\nTrying Proxy: {option}')
        proxy = urllib.request.ProxyHandler({'http': option})
    return proxy


def create_url(url, _print=False):
    """
    From the given url, produce a URL that is compatible with Github's REST API. Can handle blob or tree paths.
    """
    repo_only_url = re.compile(r"https:\/\/github\.com\/[a-z\d](?:[a-z\d]|-(?=[a-z\d])){0,38}\/[a-zA-Z0-9]+")
    re_branch = re.compile("/(tree|blob)/(.+?)/")

    # Check if the given url is a url to a GitHub repo. If it is, tell the
    # user to use 'git clone' to download it
    branch = re_branch.search(url)
    if re.match(repo_only_url,url) and branch is None:
        if _print:
            print("The given url is a complete repository. Use 'git clone' to download the repository")
        sys.exit()

    # extract the branch name from the given url (e.g master)
    if branch:
        download_dirs = url[branch.end():]
        api_url = (url[:branch.start()].replace("github.com", "api.github.com/repos", 1) +
                "/contents/" + download_dirs + "?ref=" + branch.group(2))
        return api_url, download_dirs.split('/')[-1]
    else:
        if _print:
            print("Couldn't find the repo, Pls check the URL!!!")
        sys.exit()


def download(repo_url, proxies=None, output_dir="./", flatten=False, exts=None, file_count=0, _print=False):
    """ Downloads the files and directories in repo_url. If flatten is specified, the contents of any and all
     sub-directories will be pulled upwards into the root folder. """

    # handle paths cross platform
    output_dir = Path(output_dir)

    # getting proxy from proxy list
    proxy = get_proxy(proxies, _print=_print)

    # generate the url which returns the JSON data
    api_url, download_dir = create_url(repo_url, _print=_print)

    # To handle file names.
    if not flatten:
        dir_out = Path(output_dir) / download_dir
    else:
        dir_out = Path(output_dir)

    # trying to get api response
    try:
        opener = urllib.request.build_opener(proxy)
        opener.addheaders = [('User-agent'  , 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36')]
        urllib.request.install_opener(opener)
        response = urllib.request.urlretrieve(api_url)
    except KeyboardInterrupt:
        # when CTRL+C is pressed during the execution of this script,
        # bring the cursor to the beginning, erase the current line, and dont make a new line
        if _print:
            print("Got interrupted")
        sys.exit()
    except urllib.error.HTTPError as e:
        # if e.code == 403:                      ## Api response 403 error
        #     if _print:
        #         print("API Rate limit exceeded!!!")
        #     download(repo_url, proxies, dir_out, flatten, exts=exts, file_count=file_count)  # BUG: infinit loop leads to crash if link doesn't exists
        # else:
        if _print:
            print(e)
        sys.exit()
    except:
        if _print:
            print("Failed")
        sys.exit()

    # make a directory with the name which is taken from
    # the actual repo
    try:
        os.makedirs(dir_out)
    except FileExistsError:
        pass

    with open(response[0], "r") as f:
        data = json.load(f)
        # getting the total number of files so that we
        # can use it for the output information later

        # If the data is a file, download it as one.
        if isinstance(data, dict) and data["type"] == "file":
            try:
                # download the file
                opener = urllib.request.build_opener(proxy)
                opener.addheaders = [('User-agent', 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_3) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/35.0.1916.47 Safari/537.36')]
                urllib.request.install_opener(opener)
                if exts is None:
                    urllib.request.urlretrieve(data["download_url"], Path(dir_out) / data['name'])
                    file_count += 1
                    if _print:
                        print("Downloaded: " + "{}".format(data["name"]))
                if exts is not None and os.path.splitext(data['download_url'])[1] in exts:
                    urllib.request.urlretrieve(data["download_url"], Path(dir_out) / data['name'])
                    file_count += 1
                    if _print:
                        print("Downloaded: " + "{}".format(data["name"]))
                return file_count
            except KeyboardInterrupt:
                # when CTRL+C is pressed during the execution of this script,
                # bring the cursor to the beginning, erase the current line, and dont make a new line
                if _print:
                    print("Got interrupted")
                sys.exit()
            except urllib.error.HTTPError as e:
                if e.code == 403:
                    if _print:
                        print("API Rate limit exceeded!!!")
                    download(data["html_url"], proxies, dir_out, flatten, exts=exts, file_count=file_count)
                else:
                    if _print:
                        print(e)
                sys.exit()
            except:
                if _print:
                    print("Failed")
                sys.exit()

        # going over the files in the directory
        for file in data:
            file_url = file["download_url"]
            file_name = file["name"]

            if file_url is not None:
                try:
                    # if it's a file
                    path = Path(dir_out) / file_name
                    opener = urllib.request.build_opener(proxy)
                    opener.addheaders = [('User-agent', 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_10_1) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/39.0.2171.95 Safari/537.36')]
                    urllib.request.install_opener(opener)
                    # download the file
                    if exts is None:
                        urllib.request.urlretrieve(file_url, path)
                        file_count += 1
                        if _print:
                            print("Downloaded: " + "{}".format(file_name))
                    if exts is not None and os.path.splitext(file_url)[1] in exts:
                        urllib.request.urlretrieve(file_url, path)
                        file_count += 1
                        if _print:
                            print("Downloaded: " + "{}".format(file_name), "green")

                except KeyboardInterrupt:
                    # when CTRL+C is pressed during the execution of this script,
                    # bring the cursor to the beginning, erase the current line, and dont make a new line
                    if _print:
                        print("Got interrupted")
                    sys.exit()
                except urllib.error.HTTPError as e:
                    if e.code == 403:
                        if _print:
                            print("API Rate limit exceeded!!!")
                        download(file["html_url"], proxies, dir_out, flatten, exts=exts, file_count=file_count)
                    else:
                        if _print:
                            print(e)
                    sys.exit()
                except:
                    if _print:
                        print("Failed")
                    sys.exit()
            else:
                download(file["html_url"], proxies, dir_out, flatten, exts=exts, file_count=file_count)

    return file_count


def main():

    # def dir_path(string):
    #     if os.path.isdir(string):
    #         return string
    #     else:
    #         print_text("Please enter the correct output directory path!!!", "red", in_place=True)
    #         sys.exit()


    parser = argparse.ArgumentParser(description="Download directories/folders from GitHub")
    parser.add_argument('urls', nargs="+",
                        help="List of Github directories to download.")
    parser.add_argument('--output_dir', "-d", dest="output_dir", default="./",
                        help="All directories will be downloaded to the specified directory.")

    parser.add_argument('--flatten', '-f', action="store_true",
                        help='Flatten directory structures. Do not create extra directory and download found files to'
                             ' output directory. (default to current directory if not specified)')

    parser.add_argument('--proxy', '-p', dest='proxies', default=None, help="txt file path containing http proxies.")
    parser.add_argument('--exts', '-e', nargs="+", dest='exts', default=None, help="List of File extensions which you want to download")

    args = parser.parse_args()

    flatten = args.flatten
    exts = args.exts

    proxies = args.proxies
    if proxies is not None:
        try:
            with open(proxies, 'r') as f:
                proxies = f.read().splitlines()
        except:
            print("No proxy txt file found!!!")
            proxies=None
            pass

    for url in args.urls:
        total_files = download(url, proxies, args.output_dir, flatten, exts)
        if total_files > 0:
            print("Download Complete")
        else:
            print("Files Not Found!!!")


if __name__ == "__main__":
    main()



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 22:08:29 2019

@author: qchat
"""
from typing import List
import os
import sys
import shutil
from functools import partial
from collections import OrderedDict

from qtpy import QtWidgets, QtCore, uic, QtGui

from .config import ConfigManager
from .figure import FigureManager
from .parameter import ParameterManager
from .recipe import RecipeManager
from .scan import ScanManager
from .data import DataManager
from ..icons import icons
from ..GUI_instances import openVariablesMenu, openPlotter
from ...paths import PATHS
from ...utilities import boolean, SUPPORTED_EXTENSION
from ...config import get_scanner_config, load_config, change_autolab_config


class Scanner(QtWidgets.QMainWindow):

    def __init__(self, parent: QtWidgets.QMainWindow):

        self.mainGui = parent

        # Configuration of the window
        super().__init__()
        ui_path = os.path.join(os.path.dirname(__file__), 'interface.ui')
        uic.loadUi(ui_path, self)
        self.setWindowTitle("AUTOLAB - Scanner")
        self.setWindowIcon(icons['scanner'])
        self.splitter.setSizes([500, 700])  # Set the width of the two main widgets
        self.setAcceptDrops(True)
        self.recipeDict = {}
        self._append = False  # option for import config
        self._copy_step_info = None  # used by tree (in recipe) for copy paste step

        # Loading of the different centers
        self.figureManager = FigureManager(self)
        self.scanManager = ScanManager(self)
        self.dataManager = DataManager(self)
        self.configManager = ConfigManager(self)

        # Configuration menu
        configMenu = self.menuBar.addMenu('Configuration')

        self.importAction = configMenu.addAction('Import configuration')
        self.importAction.setIcon(icons['import'])
        self.importAction.triggered.connect(self.importActionClicked)
        self.importAction.setStatusTip("Import configuration file")

        self.openRecentMenu = configMenu.addMenu('Import recent configuration')
        self.populateOpenRecent()

        configMenu.addSeparator()

        exportAction = configMenu.addAction('Export current configuration')
        exportAction.setIcon(icons['export'])
        exportAction.triggered.connect(self.exportActionClicked)
        exportAction.setStatusTip("Export current configuration file")

        # Edition menu
        editMenu = self.menuBar.addMenu('Edit')
        self.undo = editMenu.addAction('Undo')
        self.undo.setIcon(icons['undo'])
        self.undo.setShortcut(QtGui.QKeySequence("Ctrl+Z"))
        self.undo.triggered.connect(self.configManager.undoClicked)
        self.undo.setEnabled(False)
        self.undo.setStatusTip("Revert recipe changes")
        self.redo = editMenu.addAction('Redo')
        self.redo.setIcon(icons['redo'])
        self.redo.setShortcut(QtGui.QKeySequence("Ctrl+Y"))
        self.redo.triggered.connect(self.configManager.redoClicked)
        self.redo.setEnabled(False)
        self.redo.setStatusTip("Reapply recipe changes")

        guiMenu = self.menuBar.addMenu('Panels')

        plotAction = guiMenu.addAction('Plotter')
        plotAction.setIcon(icons['plotter'])
        plotAction.triggered.connect(lambda: openPlotter(has_parent=True))
        plotAction.setStatusTip('Open the plotter in another window')

        variablesMenuAction = guiMenu.addAction('Variables')
        variablesMenuAction.setIcon(icons['variables'])
        variablesMenuAction.triggered.connect(lambda: openVariablesMenu(True))
        variablesMenuAction.setStatusTip("Open the variable menu in another window")

        self.configManager.addRecipe("recipe")  # add one recipe by default
        self.configManager.undoClicked() # avoid false history
        self.setStatus("")
        self.addRecipe_pushButton.clicked.connect(
            lambda: self.configManager.addRecipe("recipe"))

        self.selectRecipe_comboBox.activated.connect(self._updateSelectParameter)

        # Save button configuration
        self.save_pushButton.clicked.connect(self.saveButtonClicked)
        self.save_pushButton.setEnabled(False)
        self.save_all_pushButton.clicked.connect(self.saveAllButtonClicked)
        self.save_all_pushButton.setEnabled(False)

        # Clear button configuration
        self.clear_pushButton.clicked.connect(self.clear)
        self.clear_pushButton.setEnabled(False)

        self.variable_x2_comboBox.hide()
        self.label_scan_2D.hide()

        for splitter in (self.splitter, self.splitterGraph):
            for i in range(splitter.count()):
                handle = splitter.handle(i)
                handle.setStyleSheet("background-color: #DDDDDD;")
                handle.installEventFilter(self)

    def eventFilter(self, obj, event):
        if event.type() == QtCore.QEvent.Enter:
            obj.setStyleSheet("background-color: #AAAAAA;")  # Hover color
        elif event.type() == QtCore.QEvent.Leave:
            obj.setStyleSheet("background-color: #DDDDDD;")  # Normal color
        return super().eventFilter(obj, event)

    def populateOpenRecent(self):
        """ https://realpython.com/python-menus-toolbars/#populating-python-menus-dynamically """
        self.openRecentMenu.clear()

        if os.path.exists(PATHS['history_config']):
            with open(PATHS['history_config'], 'r') as f:
                filenames = f.readlines()
            for filename in reversed(filenames):
                filename = filename.rstrip('\n')
                action = QtWidgets.QAction(filename, self)
                action.setIcon(icons['import'])
                action.setEnabled(os.path.exists(filename))
                action.triggered.connect(
                    partial(self.configManager.import_configPars, filename))
                self.openRecentMenu.addAction(action)

        self.openRecentMenu.addSeparator()
        action = QtWidgets.QAction('Clear list', self)
        action.setIcon(icons['remove'])
        action.triggered.connect(self.clearOpenRecent)
        self.openRecentMenu.addAction(action)

    def addOpenRecent(self, filename: str):

        if not os.path.exists(PATHS['history_config']):
            with open(PATHS['history_config'], 'w') as f:
                f.write(filename + '\n')
        else:
            with open(PATHS['history_config'], 'r') as f:
                lines = f.readlines()
            lines.append(filename)
            lines = [line.rstrip('\n')+'\n' for line in lines]
            lines = list(reversed(list(dict.fromkeys(reversed(lines)))))  # unique names
            lines = lines[-10:]
            with open(PATHS['history_config'], 'w') as f:
                f.writelines(lines)

        self.populateOpenRecent()

    def clearOpenRecent(self):
        if os.path.exists(PATHS['history_config']):
            try: os.remove(PATHS['history_config'])
            except: pass

        self.populateOpenRecent()

    def clear(self):
        """ This reset any recorded data, and the GUI accordingly """
        self.dataManager.datasets = []
        self.figureManager.clearData()
        self.figureManager.figMap.hide()
        self.figureManager.fig.show()
        self.figureManager.setLabel("x", " ")
        self.figureManager.setLabel("y", " ")
        self.frameAxis.show()
        self.variable_x_comboBox.clear()
        self.variable_x2_comboBox.clear()
        self.variable_y_comboBox.clear()
        self.variable_x2_checkBox.show()
        self.data_comboBox.clear()
        self.data_comboBox.hide()
        self.save_pushButton.setEnabled(False)
        self.save_all_pushButton.setEnabled(False)
        self.clear_pushButton.setEnabled(False)
        self.save_pushButton.setText('Save scan1')
        self.progressBar.setValue(0)
        self.progressBar.setStyleSheet("")
        self.displayScanData_pushButton.hide()
        self.sendScanData_pushButton.hide()
        self.dataframe_comboBox.clear()
        self.dataframe_comboBox.addItems(["Scan"])
        self.dataframe_comboBox.hide()
        self.scan_recipe_comboBox.setCurrentIndex(0)
        self.scan_recipe_comboBox.hide()
        self.refresh_widget(self.clear_pushButton)

    @staticmethod
    def refresh_widget(widget: QtWidgets.QWidget):
        """ Avoid intermediate disabled state when badly refreshed """
        widget.style().unpolish(widget)
        widget.style().polish(widget)
        widget.update()

    def _addRecipe(self, recipe_name: str):
        """ Adds recipe to managers. Called by configManager """
        self._update_recipe_combobox()  # recreate all and display first index
        self.selectRecipe_comboBox.setCurrentIndex(
            self.selectRecipe_comboBox.count()-1)  # display last index

        self.recipeDict[recipe_name] = {}  # order of creation matter
        self.recipeDict[recipe_name]['recipeManager'] = RecipeManager(self, recipe_name)
        self.recipeDict[recipe_name]['parameterManager'] = OrderedDict()

        for parameter in self.configManager.parameterList(recipe_name):
            self._addParameter(recipe_name, parameter['name'])

    def _removeRecipe(self, recipe_name: str):  # order of creation matter
        """ Removes recipe from managers. Called by configManager and self """
        test = self.recipeDict.pop(recipe_name)
        test['recipeManager']._removeWidget()
        self._update_recipe_combobox()
        self._updateSelectParameter()

    def _activateRecipe(self, recipe_name: str, state: bool):
        """ Activates/Deactivates an existing recipe. Called by configManager and recipeManager """
        active = bool(state)
        self._update_recipe_combobox()
        self.recipeDict[recipe_name]['recipeManager']._activateTree(active)

    def _update_recipe_combobox(self):
        """ Shows recipe combobox if multi recipes else hide """
        prev_index = self.selectRecipe_comboBox.currentIndex()

        self.selectRecipe_comboBox.clear()
        self.selectRecipe_comboBox.addItems(self.configManager.recipeNameList())

        new_index = min(prev_index, self.selectRecipe_comboBox.count()-1)
        self.selectRecipe_comboBox.setCurrentIndex(new_index)

        dataSet_id = len(self.configManager.recipeNameList())
        if dataSet_id > 1:
            self.selectRecipe_comboBox.show()
            self.label_selectRecipeParameter.show()
        else:
            self.selectRecipe_comboBox.hide()
            # OPTIMIZE: seperate condition for label
            recipe_name = self.selectRecipe_comboBox.currentText()
            if (recipe_name != ""
                    and len(self.configManager.parameterList(recipe_name)) > 1):
                self.selectParameter_comboBox.show()
            else:
                self.label_selectRecipeParameter.hide()

    def _clearRecipe(self):
        """ Clears recipes from managers. Called by configManager """
        for recipe_name in list(self.recipeDict):
            self._removeRecipe(recipe_name)

    def _addParameter(self, recipe_name: str, param_name: str):
        """ Adds parameter to managers. Called by configManager and self """
        new_ParameterManager = ParameterManager(self, recipe_name, param_name)
        self.recipeDict[recipe_name]['parameterManager'][param_name] = new_ParameterManager

        layoutAll = self.recipeDict[recipe_name]['recipeManager']._layoutAll
        layoutAll.insertWidget(
            layoutAll.count()-1, new_ParameterManager.mainFrame, stretch=0)

        self._updateSelectParameter()
        self.selectParameter_comboBox.setCurrentIndex(
            self.selectParameter_comboBox.count()-1)

    def _removeParameter(self, recipe_name: str, param_name: str):
        """ Removes parameter from managers. Called by configManager """
        test = self.recipeDict[recipe_name]['parameterManager'].pop(param_name)
        test._removeWidget()

        self._updateSelectParameter()

    def _updateSelectParameter(self):
        """ Updates selectParameter_comboBox. Called by configManager and self """
        recipe_name = self.selectRecipe_comboBox.currentText()

        prev_index = self.selectParameter_comboBox.currentIndex()
        if prev_index == -1: prev_index = 0

        self.selectParameter_comboBox.clear()
        if recipe_name != "":
            self.selectParameter_comboBox.addItems(
                self.configManager.parameterNameList(recipe_name))
            self.selectParameter_comboBox.setCurrentIndex(prev_index)

        if self.selectParameter_comboBox.currentText() == "":
            self.selectParameter_comboBox.setCurrentIndex(
                self.selectParameter_comboBox.count()-1)

        #Shows parameter combobox if multi parameters else hide
        if (recipe_name != ""
                and len(self.configManager.parameterList(recipe_name)) > 1):
            self.selectParameter_comboBox.show()
            self.label_selectRecipeParameter.show()
        else:
            self.selectParameter_comboBox.hide()
            if not self.selectRecipe_comboBox.isVisible():
                self.label_selectRecipeParameter.hide()

    def _refreshParameterRange(self, recipe_name: str, param_name: str,
                               newName: str = None):
        """ Updates parameterManager with new parameter name """
        recipeDictParam = self.recipeDict[recipe_name]['parameterManager']

        if newName is None:
            recipeDictParam[param_name].refresh()
        else:
            if param_name in recipeDictParam:
                recipeDictParam[newName] = recipeDictParam.pop(param_name)
                recipeDictParam[newName].changeName(newName)
                recipeDictParam[newName].refresh()
            else:
                print(f"Error: Can't refresh parameter '{param_name}', not found in recipeDictParam '{recipeDictParam}'",
                      file=sys.stderr)

        self._updateSelectParameter()

    def _refreshRecipe(self, recipe_name: str):
        self.recipeDict[recipe_name]['recipeManager'].refresh()

    def _resetRecipe(self):
        """ Resets recipe """
        self._clearRecipe()  # before everything to have access to recipe and del it

        for recipe_name in self.configManager.recipeNameList():
            self._addRecipe(recipe_name)
            for parameterManager in self.recipeDict[recipe_name]['parameterManager'].values():
                parameterManager.refresh()
            self._refreshRecipe(recipe_name)

    def importActionClicked(self):
        """ Prompts the user for a configuration filename,
        and import the current scan configuration from it """

        class ImportDialog(QtWidgets.QDialog):

            def __init__(self, parent: QtWidgets.QMainWindow, append: bool):

                super().__init__(parent)
                self.setWindowTitle("Import AUTOLAB configuration file")
                self.setWindowModality(QtCore.Qt.ApplicationModal)  # this block GUI interaction (easier than checking every interaction possible to avoid bugs if change recipe or have multiple dialogs)

                self.append = append

                urls = [
                    QtCore.QUrl.fromLocalFile(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.HomeLocation)),
                    QtCore.QUrl.fromLocalFile(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.DesktopLocation)),
                    QtCore.QUrl.fromLocalFile(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.DocumentsLocation)),
                    QtCore.QUrl.fromLocalFile(QtCore.QStandardPaths.writableLocation(QtCore.QStandardPaths.DownloadLocation)),
                    QtCore.QUrl.fromLocalFile(os.environ['TEMP']),
                ]

                layout = QtWidgets.QVBoxLayout(self)
                self.setLayout(layout)

                file_dialog = QtWidgets.QFileDialog(self, QtCore.Qt.Widget)
                file_dialog.setOption(QtWidgets.QFileDialog.DontUseNativeDialog)
                file_dialog.setSidebarUrls(urls)
                file_dialog.setWindowFlags(file_dialog.windowFlags() & ~QtCore.Qt.Dialog)
                file_dialog.setDirectory(PATHS['last_folder'])
                file_dialog.setNameFilters(["AUTOLAB configuration file (*.conf)", "Any Files (*)"])
                layout.addWidget(file_dialog)

                appendCheck = QtWidgets.QCheckBox('Append', self)
                appendCheck.setChecked(append)
                appendCheck.stateChanged.connect(self.appendCheckChanged)
                layout.addWidget(appendCheck)

                self.exec_ = file_dialog.exec_
                self.selectedFiles = file_dialog.selectedFiles

            def appendCheckChanged(self, event):
                self.append = event

            def closeEvent(self, event):
                for children in self.findChildren(QtWidgets.QWidget):
                    children.deleteLater()

                super().closeEvent(event)

        main_dialog = ImportDialog(self, self._append)
        main_dialog.show()

        once_or_append = True
        while once_or_append:
            if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                filenames = main_dialog.selectedFiles()

                self._append = main_dialog.append
                once_or_append = self._append and len(filenames) != 0

                for filename in filenames:
                    if filename != '':
                        self.configManager.import_configPars(
                            filename, append=self._append)
            else:
                once_or_append = False

        main_dialog.deleteLater()

    def exportActionClicked(self):
        """ Prompts the user for a configuration file path,
        and export the current scan configuration in it """
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export AUTOLAB configuration file",
            os.path.join(PATHS['last_folder'], 'config.conf'),
            "AUTOLAB configuration file (*.conf);;All Files (*)")[0]

        if filename != '':
            path = os.path.dirname(filename)
            PATHS['last_folder'] = path

            try:
                self.configManager.export(filename)
                self.setStatus(
                    f"Current configuration successfully saved at {filename}",
                    5000)
            except Exception as e:
                self.setStatus(f"An error occured: {str(e)}", 10000, False)
            else:
                self.addOpenRecent(filename)

    def saveAllButtonClicked(self):
        self._saveData(self.dataManager.datasets)

    def saveButtonClicked(self):
        self._saveData([self.dataManager.getLastSelectedDataset()])

    def _saveData(self, all_data: List[dict]):
        """ This function is called when the save button is clicked.
        It asks a path and starts the procedure to save the data """
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self,  caption="Save data",
            directory=PATHS['last_folder'],
            filter=SUPPORTED_EXTENSION)[0]
        path = os.path.dirname(filename)

        save_folder, extension = os.path.splitext(filename)

        if path != '':
            PATHS['last_folder'] = path
            self.setStatus('Saving data...', 5000)

            for scanset in all_data:
                i = self.dataManager.datasets.index(scanset)
                scan_name = f'scan{i+1}'

                if len(all_data) == 1:
                    scan_filename = save_folder
                    new_configname = f'{save_folder}.conf'
                else:
                    scan_filename = f'{save_folder}_{scan_name}'
                    new_configname = f'{save_folder}_{scan_name}.conf'

                for recipe_name in scanset:
                    dataset = scanset[recipe_name]

                    if len(scanset) == 1:
                        filename_recipe = f'{scan_filename}{extension}'
                    else:
                        filename_recipe = f'{scan_filename}_{recipe_name}{extension}'
                    dataset.save(filename_recipe)

                scanset.saved = True
                scanner_config = get_scanner_config()
                save_config = boolean(scanner_config["save_config"])

                if save_config:
                    config_name = os.path.join(
                        os.path.dirname(dataset.folder_dataset_temp), 'config.conf')

                    if os.path.exists(config_name):
                        shutil.copy(config_name, new_configname)
                    else:
                        if scanset is not self.dataManager.getLastDataset():
                            print("Warning: Can't find config for this dataset, save latest config instead",
                                  file=sys.stderr)
                        self.configManager.export(new_configname)  # BUG: it saves latest config instead of dataset config because no record available of previous config. (I did try to put back self.config to dataset but config changes with new dataset (copy doesn't help and deepcopy not possible)

                    self.addOpenRecent(new_configname)

                if len(all_data) == 1 and boolean(scanner_config["save_figure"]):
                        self.figureManager.save(filename)

            if len(all_data) == 1:
                self.setStatus(
                    f'{scan_name} successfully saved in {filename}', 5000)
            else:
                self.setStatus(
                    f'All scans successfully saved as {save_folder}_[...].txt', 5000)

    def dropEvent(self, event):
        """ Imports config file if event has url of a file """
        filename = event.mimeData().urls()[0].toLocalFile()
        self.configManager.import_configPars(filename)

        qwidget_children = self.findChildren(QtWidgets.QWidget)
        for widget in qwidget_children:
            widget.setGraphicsEffect(None)

    def dragEnterEvent(self, event):
        """ Only accept config file (url) """
        if event.mimeData().hasUrls():
            event.accept()

            qwidget_children = self.findChildren(QtWidgets.QWidget)
            for widget in qwidget_children:
                shadow = QtWidgets.QGraphicsColorizeEffect()
                shadow.setColor(QtGui.QColor(20,20,20))
                shadow.setStrength(1)
                widget.setGraphicsEffect(shadow)
        else:
            event.ignore()

    def dragLeaveEvent(self, event):
        qwidget_children = self.findChildren(QtWidgets.QWidget)
        for widget in qwidget_children:
            widget.setGraphicsEffect(None)

    def closeEvent(self, event):
        """ Does some steps before the window is really killed """
        # Stop ongoing scan
        if self.scanManager.isStarted():
            self.scanManager.stop()

        # Stop datamanager timer
        self.dataManager.timer.stop()

        scanner_config = get_scanner_config()
        ask_close = boolean(scanner_config["ask_close"])
        if ask_close and not all([scanset.saved for scanset in self.dataManager.datasets]):
            msg_box = QtWidgets.QMessageBox(self)
            msg_box.setWindowTitle("Scanner")
            msg_box.setText("Some data hasn't been saved, close scanner anyway?")
            msg_box.setStandardButtons(QtWidgets.QMessageBox.Yes
                                       | QtWidgets.QMessageBox.No)

            checkbox = QtWidgets.QCheckBox("Don't ask again")
            msg_box.setCheckBox(checkbox)

            msg_box.show()
            res = msg_box.exec_()

            autolab_config = load_config('autolab_config')
            autolab_config['scanner']['ask_close'] = str(not checkbox.isChecked())
            change_autolab_config(autolab_config)

            if res == QtWidgets.QMessageBox.No:
                event.ignore()  # Prevent the window from closing
                return None

        self.mainGui.clearScanner()

        for recipe in self.recipeDict.values():
            for parameterManager in recipe['parameterManager'].values():
                parameterManager.close()

        self.figureManager.close()

        # Remove scan variables from VARIABLES
        try: self.configManager.updateVariableConfig([])
        except: pass

        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().closeEvent(event)

    def setStatus(self, message: str, timeout: int = 0, stdout: bool = True):
        """ Modifies displayed message in status bar and adds error message to logger """
        self.statusBar.showMessage(message, timeout)
        if not stdout: print(message, file=sys.stderr)



# -*- coding: utf-8 -*-
"""
Created on Mon Aug  5 14:32:41 2024

@author: Jonathan
"""

import sys
import platform

import numpy as np
import pandas as pd
import qtpy
from qtpy import QtCore, QtWidgets
import pyqtgraph as pg

from .GUI_instances import clearAbout
from .icons import icons

from ..web import project_url, drivers_url, doc_url
from ... import __version__


def get_versions() -> dict:
    """Information about Autolab versions """

    # Based on Spyder about.py (https://github.com/spyder-ide/spyder/blob/3ce32d6307302a93957594569176bc84d9c1612e/spyder/plugins/application/widgets/about.py#L40)
    versions = {
        'autolab': __version__,
        'python': platform.python_version(),  # "2.7.3"
        'bitness': 64 if sys.maxsize > 2**32 else 32,
        'qt_api': qtpy.API_NAME,      # PyQt5
        'qt_api_ver': (qtpy.PYSIDE_VERSION if 'pyside' in qtpy.API
                       else qtpy.PYQT_VERSION),
        'system': platform.system(),   # Linux, Windows, ...
        'release': platform.release(),  # XP, 10.6, 2.2.0, etc.
        'pyqtgraph': pg.__version__,
        'numpy': np.__version__,
        'pandas': pd.__version__,
    }
    if sys.platform == 'darwin':
        versions.update(system='macOS', release=platform.mac_ver()[0])

    return versions


class AboutWindow(QtWidgets.QMainWindow):

    def __init__(self, parent: QtWidgets.QMainWindow = None):

        super().__init__()
        self.mainGui = parent
        self.setWindowTitle('AUTOLAB - About')
        self.setWindowIcon(icons['autolab'])

        self.init_ui()

        self.adjustSize()

        # Don't want to have about windows taking the full screen
        self.setWindowFlags(QtCore.Qt.Window
                            | QtCore.Qt.WindowMinimizeButtonHint
                            | QtCore.Qt.WindowCloseButtonHint
                            | QtCore.Qt.WindowTitleHint)

    def init_ui(self):
        versions = get_versions()

        # Main layout creation
        layoutWindow = QtWidgets.QVBoxLayout()
        layoutTab = QtWidgets.QHBoxLayout()
        layoutWindow.addLayout(layoutTab)
        layoutWindow.setContentsMargins(0,0,0,0)
        layoutWindow.setSpacing(0)

        centralWidget = QtWidgets.QWidget()
        centralWidget.setLayout(layoutWindow)
        self.setCentralWidget(centralWidget)

        frameOverview = QtWidgets.QFrame()
        layoutOverview = QtWidgets.QVBoxLayout(frameOverview)
        layoutOverview.setAlignment(QtCore.Qt.AlignTop)

        frameLegal = QtWidgets.QFrame()
        layoutLegal = QtWidgets.QVBoxLayout(frameLegal)
        layoutLegal.setAlignment(QtCore.Qt.AlignTop)

        tab = QtWidgets.QTabWidget(self)
        tab.addTab(frameOverview, 'Overview')
        tab.addTab(frameLegal, 'Legal')

        label_pic = QtWidgets.QLabel()
        label_pic.setPixmap(icons['autolab-pixmap'])

        label_autolab = QtWidgets.QLabel(f"<h2>Autolab {versions['autolab']}</h2>")
        label_autolab.setAlignment(QtCore.Qt.AlignCenter)

        frameIcon = QtWidgets.QFrame()
        layoutIcon = QtWidgets.QVBoxLayout(frameIcon)
        layoutIcon.addWidget(label_pic)
        layoutIcon.addWidget(label_autolab)
        layoutIcon.addStretch()

        layoutTab.addWidget(frameIcon)
        layoutTab.addWidget(tab)

        label_versions = QtWidgets.QLabel(
            f"""
            <h1>Autolab</h1>

            <h3>Python package for scientific experiments automation</h3>

            <p>
            {versions['system']} {versions['release']}
            <br>
            Python {versions['python']} - {versions['bitness']}-bit
            <br>
            {versions['qt_api']} {versions['qt_api_ver']} |
            PyQtGraph {versions['pyqtgraph']} |
            Numpy {versions['numpy']} |
            Pandas {versions['pandas']}
            </p>

            <p>
            <a href="{project_url}">Project</a> |
            <a href="{drivers_url}">Drivers</a> |
            <a href="{doc_url}"> Documentation</a>
            </p>
            """
        )
        label_versions.setOpenExternalLinks(True)
        label_versions.setWordWrap(True)

        layoutOverview.addWidget(label_versions)

        label_legal = QtWidgets.QLabel(
            f"""
            <p>
            Created by <b>Quentin Chateiller</b>, Python drivers originally from
            Quentin Chateiller and <b>Bruno Garbin</b>, for the C2N-CNRS
            (Center for Nanosciences and Nanotechnologies, Palaiseau, France)
            ToniQ team.
            <br>
            Project continued by <b>Jonathan Peltier</b>, for the C2N-CNRS
            Minaphot team and <b>Mathieu Jeannin</b>, for the C2N-CNRS
            Odin team.
            <br>
            <br>
            Distributed under the terms of the
            <a href="{project_url}/blob/master/LICENSE">GPL-3.0 licence</a>
            </p>"""
        )
        label_legal.setOpenExternalLinks(True)
        label_legal.setWordWrap(True)
        layoutLegal.addWidget(label_legal)

    def closeEvent(self, event):
        """ Does some steps before the window is really killed """
        clearAbout()

        if not self.mainGui:
            QtWidgets.QApplication.quit()  # close the about app



# -*- coding: utf-8 -*-
"""
Created on Thu Nov  9 10:58:47 2023

@author: Jonathan
"""
import pandas as pd
from qtpy import QtCore, QtWidgets
import pyqtgraph as pg


class DisplayValues(QtWidgets.QWidget):

    def __init__(self, name: str, size: QtCore.QSize = (250, 400)):
        """ Create a QWidget displaying the dataFrame input to the refresh method.
        size is of type QtCore.QSize or tuple of int """

        if isinstance(size, (tuple, list)): size = QtCore.QSize(*size)

        self.active = False

        super().__init__()

        self.setWindowTitle(name)
        self.resize(size)
        # I choose pg.TableWidget to have more features at the cost of slower execution than QTableView+TableModel
        # self.tableView = QtWidgets.QTableView()  # basic (can't save/copy data)
        self.tableView = pg.TableWidget()  # slow
        self.tableView.setFormat('%.16g')

        layoutWindow = QtWidgets.QVBoxLayout()
        layoutWindow.addWidget(self.tableView)
        self.setLayout(layoutWindow)

    def refresh(self, data: pd.DataFrame):
        """ data is pd.DataFrame """
        # tableModel = TableModel(data)
        # self.tableView.setModel(tableModel)
        self.tableView.setData(data.to_dict('records'))

    def show(self):
        if not self.active:
            super().show()
            self.active = True
        else:
            self.setWindowState(self.windowState() & ~QtCore.Qt.WindowMinimized | QtCore.Qt.WindowActive)
            self.activateWindow()

    def closeEvent(self, event):
        self.active = False

    def close(self):
        for children in self.findChildren(QtWidgets.QWidget):
            children.deleteLater()

        super().close()


# OBSOLETE
class TableModel(QtCore.QAbstractTableModel):
    "From https://www.pythonguis.com/tutorials/pyqt6-qtableview-modelviews-numpy-pandas/"
    def __init__(self, data: pd.DataFrame):
        super().__init__()
        self._data = data

    def data(self, index, role):
        if hasattr(QtCore.Qt.ItemDataRole, 'DisplayRole'):
            role_check = QtCore.Qt.ItemDataRole.DisplayRole
        else:
            role_check = 0  # Compatibility py3.6
        if role == role_check:
            value = self._data.iloc[index.row(), index.column()]
            return str(value)

    def rowCount(self, index):
        return self._data.shape[0]

    def columnCount(self, index):
        return self._data.shape[1]

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if hasattr(QtCore.Qt.ItemDataRole, 'DisplayRole'):
            role_check = QtCore.Qt.ItemDataRole.DisplayRole
        else:
            role_check = 0
        if role == role_check:
            if orientation == QtCore.Qt.Horizontal:
                return str(self._data.columns[section])

            if orientation == QtCore.Qt.Vertical:
                return str(self._data.index[section])



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:22:35 2019

@author: qchat
"""

import os

import numpy as np
import pyqtgraph as pg
import pyqtgraph.exporters  # Needed for pg.exporters.ImageExporter
from qtpy import QtWidgets

from ..GUI_utilities import pyqtgraph_fig_ax, pyqtgraph_image
from ...config import get_monitor_config
from ...utilities import boolean


class FigureManager:

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui

        # Import Autolab config
        monitor_config = get_monitor_config()
        self.precision = int(monitor_config['precision'])
        self.do_save_figure = boolean(monitor_config['save_figure'])

        # Configure and initialize the figure in the GUI
        self.fig, self.ax = pyqtgraph_fig_ax()
        self.gui.graph.addWidget(self.fig)
        self.figMap, widget = pyqtgraph_image()
        self.gui.graph.addWidget(widget)
        self.figMap.hide()

        self.setLabel('x', self.gui.xlabel)
        self.setLabel('y', self.gui.ylabel)

        self.plot = self.ax.plot([], [], symbol='x', pen='r', symbolPen='r',
                                 symbolSize=10, symbolBrush='r')
        self.plot_mean = self.ax.plot([], [], pen=pg.mkPen(
            color=pg.getConfigOption("foreground"), width=2, style=pg.QtCore.Qt.DashLine))
        self.plot_min = self.ax.plot([], [], pen=pg.mkPen(color=pg.getConfigOption("foreground"), width=2))
        self.plot_max = self.ax.plot([], [], pen=pg.mkPen(color=pg.getConfigOption("foreground"), width=2))
        self.ymin = None
        self.ymax = None

    # PLOT DATA
    ###########################################################################

    def update(self, xlist: list, ylist: list) -> None:
        """ This function update the figure in the GUI """
        if xlist is None: # image
            self.figMap.setImage(ylist, autoRange=False, autoLevels=False, autoHistogramRange=False)
            if self.fig.isVisible():
                self.fig.hide()
                self.gui.min_checkBox.hide()
                self.gui.mean_checkBox.hide()
                self.gui.max_checkBox.hide()
                self.figMap.show()
                self.figMap.autoLevels()  # Only set levels for first acquisition (others are autoLevels disabled)
            return None

        if not self.fig.isVisible():
            self.fig.show()
            self.gui.min_checkBox.show()
            self.gui.mean_checkBox.show()
            self.gui.max_checkBox.show()
            self.figMap.hide()

        # Data retrieval
        try:
            self.plot.setData(xlist, ylist)
        except Exception as e:
            self.gui.setStatus(f'Error: {e}', 10000, False)
            if not self.gui.monitorManager.isPaused():
                self.gui.pauseButtonClicked()
            return None

        xlist, ylist = self.plot.getData()

        if xlist is None or ylist is None or len(xlist) == 0 or len(ylist) == 0:
            return None

        xmin = min(xlist)
        xmax = max(xlist)
        ymin = min(ylist)
        ymax = max(ylist)

        if self.ymin is None: self.ymin = ymin
        if self.ymax is None: self.ymax = ymax

        if ymin < self.ymin: self.ymin = ymin
        if ymax > self.ymax: self.ymax = ymax

        # Mean update
        if self.gui.mean_checkBox.isChecked():
            ymean = np.mean(ylist)
            self.plot_mean.setData([xmin, xmax], [ymean, ymean])

        # Min update
        if self.gui.min_checkBox.isChecked():
            self.plot_min.setData([xmin, xmax], [self.ymin, self.ymin])

        # Max update
        if self.gui.max_checkBox.isChecked():
            self.plot_max.setData([xmin, xmax], [self.ymax, self.ymax])

        # Figure finalization
        if len(ylist) >= 1:
            self.gui.dataDisplay.setText(f'{ylist[-1]:.{self.precision}g}')

            width = self.gui.dataDisplay.geometry().width()
            # height = self.gui.dataDisplay.geometry().height()
            # limit = min((width, height))
            new_size = max(width/8, 10)
            new_size = min(new_size, 25)

            font = self.gui.dataDisplay.font()
            font.setPointSize(int(new_size))
            self.gui.dataDisplay.setFont(font)

    def setLabel(self, axe: str, value: str):
        """ This function changes the label of the given axis """
        axes = {'x':'bottom', 'y':'left'}
        if value == '': value = ' '
        self.ax.setLabel(axes[axe], value, **{'color': pg.getConfigOption("foreground"),
                                              'font-size': '12pt'})

    def clear(self):
        self.ymin = None
        self.ymax = None
        self.update([],[])
        self.gui.dataDisplay.clear()

    def save(self, filename: str):
        """ This function save the figure with the provided filename """
        if self.do_save_figure:
            raw_name, extension = os.path.splitext(filename)
            new_filename = raw_name+".png"

            if not self.fig.isHidden():
                exporter = pg.exporters.ImageExporter(self.ax)
                exporter.export(new_filename)
            else:
                self.figMap.export(new_filename)



# -*- coding: utf-8 -*-
"""
Created on Oct 2022

@author: jonathan based on qchat
"""
import os

import pyqtgraph as pg
import pyqtgraph.exporters  # Needed for pg.exporters.ImageExporter

from qtpy import QtWidgets

from ..GUI_utilities import pyqtgraph_fig_ax


class FigureManager:

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui
        self.curves = []

        # Configure and initialize the figure in the GUI
        self.fig, self.ax = pyqtgraph_fig_ax()
        self.gui.graph.addWidget(self.fig)

        # Number of traces
        self.nbtraces = 10

    def start(self, new_dataset=None):
        """ This function display data and ajust buttons """
        try:
            result_names = [dataset.name for dataset in self.gui.dataManager.datasets]
            all_items = [self.gui.data_comboBox.itemText(i) for i in range(self.gui.data_comboBox.count())]

            index = self.gui.data_comboBox.currentIndex()

            if result_names != all_items:  # only refresh if change labels, to avoid gui refresh that prevent user to click on combobox
                self.gui.data_comboBox.clear()
                self.gui.data_comboBox.addItems(result_names)  # slow (0.25s)

            if new_dataset is None:
                if (index + 1) > len(result_names) or index == -1: index = 0
                self.gui.data_comboBox.setCurrentIndex(index)
            else:
                index = self.gui.data_comboBox.findText(new_dataset.name)
                self.gui.data_comboBox.setCurrentIndex(index)

            data_name = self.gui.data_comboBox.currentText() # trigger the currentIndexChanged event but don't trigger activated

            self.gui.dataManager.updateDisplayableResults()

            self.gui.save_pushButton.setEnabled(True)
            self.gui.clear_pushButton.setEnabled(True)
            self.gui.clear_all_pushButton.setEnabled(True)
            self.gui.openButton.setEnabled(True)

            self.gui.setStatus(f'Data {data_name} plotted!', 5000)

        except Exception as e:
            self.gui.setStatus(
                f'ERROR The data cannot be plotted with the given dataset: {e}',
                10000, False)

    # AXE LABEL
    ###########################################################################

    def setLabel(self, axe: str, value: str):
        """ This function changes the label of the given axis """
        axes = {'x': 'bottom', 'y': 'left'}
        if value == '': value = ' '
        self.ax.setLabel(axes[axe], value, **{'color': pg.getConfigOption("foreground"),
                                              'font-size': '12pt'})


    # PLOT DATA
    ###########################################################################

    def clearData(self):
        """ This function removes any plotted curves """
        for curve in self.curves:
            self.ax.removeItem(curve)
        self.curves = []

    def reloadData(self):
        ''' This function removes any plotted curves and reload all required curves from
        data available in the data manager'''
        # Remove all curves
        self.clearData()

        # Get current displayed result
        variable_x = self.gui.variable_x_comboBox.currentText()
        variable_y = self.gui.variable_y_comboBox.currentText()
        data_id = int(self.gui.data_comboBox.currentIndex()) + 1
        data_len = len(self.gui.dataManager.datasets)
        selectedData = data_len - data_id
        # Label update
        self.setLabel('x',variable_x)
        self.setLabel('y',variable_y)

        # Load the last results data
        try :
            # OPTIMIZE: currently load all data and plot more than self.nbtraces if in middle
            # Should change to only load nbtraces and plot nbtraces
            data = self.gui.dataManager.getData(
                data_len, [variable_x,variable_y], selectedData=0)
            # data = self.gui.dataManager.getData(self.nbtraces,[variable_x,variable_y], selectedData=selectedData)
        except:
            data = None

        # Plot them
        if data is not None:

            for i in range(len(data)):
                if i != (data_id - 1):
                    # Data
                    subdata = data[i]
                    if subdata is None:
                        continue

                    subdata = subdata.astype(float)
                    x = subdata.loc[:,variable_x]
                    y = subdata.loc[:,variable_y]

                    # Apprearance:
                    color = pg.getConfigOption("foreground")
                    alpha = (self.nbtraces - abs(data_id - 1 - i)) / self.nbtraces
                    if alpha < 0: alpha = 0

                    # Plot
                    # OPTIMIZE: keep previous style to avoid overwriting it everytime

                    if i < (data_id - 1):
                        if len(x) > 300:
                            curve = self.ax.plot(x, y, pen=color)
                            curve.setAlpha(alpha, False)
                        else:
                            curve = self.ax.plot(
                                x, y, symbol='x', symbolPen=color,
                                symbolSize=10, pen=color, symbolBrush=color)
                            curve.setAlpha(alpha, False)
                    elif i > (data_id - 1):
                        if len(x) > 300:
                            curve = self.ax.plot(
                                x, y, pen=pg.mkPen(color=color,
                                                   style=pg.QtCore.Qt.DashLine))
                            curve.setAlpha(alpha, False)
                        else:
                            curve = self.ax.plot(
                                x, y, symbol='x', symbolPen=color, symbolSize=10,
                                pen=pg.mkPen(color=color, style=pg.QtCore.Qt.DashLine),
                                symbolBrush=color)
                            curve.setAlpha(alpha, False)
                    self.curves.append(curve)

            # Data
            i = (data_id - 1)
            subdata = data[i]

            if subdata is not None:
                subdata = subdata.astype(float)
                x = subdata.loc[:, variable_x]
                y = subdata.loc[:, variable_y]

                # Apprearance:
                color = '#1f77b4'
                alpha = 1

                # Plot
                if len(x) > 300:
                    curve = self.ax.plot(x, y, pen=color)
                    curve.setAlpha(alpha, False)
                else:
                    curve = self.ax.plot(
                        x, y, symbol='x', symbolPen=color, symbolSize=10,
                        pen=color, symbolBrush=color)
                    curve.setAlpha(alpha, False)
                self.curves.append(curve)

            self.gui.plugin_refresh()


    # SAVE FIGURE
    ###########################################################################

    def save(self, filename: str):
        """ This function save the figure with the provided filename """

        raw_name, extension = os.path.splitext(filename)
        new_filename = raw_name+".png"
        exporter = pg.exporters.ImageExporter(self.ax)
        exporter.export(new_filename)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 17:11:57 2019

@author: qchat
"""

import os

AUTOLAB_FOLDER = os.path.dirname(os.path.dirname(__file__))

VERSION = os.path.join(AUTOLAB_FOLDER, 'version.txt')

LAST_FOLDER = os.path.expanduser('~')
USER_FOLDER = os.path.join(os.path.expanduser('~'), 'autolab')
DEVICES_CONFIG = os.path.join(USER_FOLDER, 'devices_config.ini')
AUTOLAB_CONFIG = os.path.join(USER_FOLDER, 'autolab_config.ini')
PLOTTER_CONFIG = os.path.join(USER_FOLDER, 'plotter_config.ini')
HISTORY_CONFIG = os.path.join(USER_FOLDER, '.history_config.txt')

# Drivers locations
DRIVERS = os.path.join(USER_FOLDER, 'drivers')
DRIVER_LEGACY = {'official': os.path.join(AUTOLAB_FOLDER, 'drivers'),
                 'local': os.path.join(USER_FOLDER, 'local_drivers')}
# can add paths in autolab_config.ini [extra_driver_path]
DRIVER_SOURCES = {'official': os.path.join(DRIVERS, 'official'),
                  'local': os.path.join(DRIVERS, 'local')}

# Driver repository
# can add paths in autolab_config.ini [extra_driver_url_repo]
# format is {'path to install': 'url to download'}
DRIVER_REPOSITORY = {DRIVER_SOURCES['official']: 'https://github.com/autolab-project/autolab-drivers'}

PATHS = {'autolab_folder': AUTOLAB_FOLDER, 'version': VERSION,
         'user_folder': USER_FOLDER, 'drivers': DRIVERS,
         'devices_config': DEVICES_CONFIG, 'autolab_config': AUTOLAB_CONFIG,
         'plotter_config': PLOTTER_CONFIG, 'history_config': HISTORY_CONFIG,
         'last_folder': LAST_FOLDER}

# Storage of the drivers paths
DRIVERS_PATHS = {}



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:10:31 2019

@author: qchat
"""

from collections import OrderedDict
from queue import Queue
import os
import shutil
import tempfile
import sys
import random
from typing import List, Union

import numpy as np
import pandas as pd
from qtpy import QtCore, QtWidgets

from ...config import get_scanner_config
from ...utilities import boolean, create_array, data_to_dataframe
from ...variables import has_eval, eval_safely


class DataManager:
    """ Manage data from a scan """

    def __init__(self, gui: QtWidgets.QMainWindow):

        self.gui = gui
        self.datasets = []
        self.queue = Queue()

        scanner_config = get_scanner_config()
        self.save_temp = boolean(scanner_config["save_temp"])

        # Timer
        self.timer = QtCore.QTimer(self.gui)
        self.timer.setInterval(33) #30fps
        self.timer.timeout.connect(self.sync)

    def getData(self, nbDataset: int, var_list: List[str],
                selectedData: int = 0, data_name: str = "Scan",
                filter_condition: List[dict] = []) -> List[pd.DataFrame]:
        """ Returns the required data """
        dataList = []
        recipe_name = self.gui.scan_recipe_comboBox.currentText()

        for i in range(selectedData, nbDataset+selectedData):
            if i < len(self.datasets):
                scanset = self.datasets[-(i+1)]
                if recipe_name not in scanset: continue
                if not scanset.display: continue
                dataset = scanset[recipe_name]
                data = None

                if data_name == "Scan":
                    try:
                        data = dataset.getData(var_list, data_name=data_name,
                                               filter_condition=filter_condition)  # OPTIMIZE: Currently can't recover dataset if error before end of first recipe loop
                    except KeyError:
                        pass  # this occurs when plot variable from scani that is not in scanj
                    except Exception as e:
                        self.gui.setStatus(
                            f"Scan warning: Can't plot Scan{len(self.datasets)-i}: {e}",
                            10000, False)
                    dataList.append(data)
                elif dataset.data_arrays.get(data_name) is not None:
                    dataList2 = []

                    try:
                        ids = dataset.getData(['id'], data_name='Scan',
                                              filter_condition=filter_condition)
                        ids = ids['id'].values - 1
                    except KeyError:
                        ids = []

                    for index in ids:
                        try:
                            data = dataset.getData(
                                var_list, data_name=data_name, dataID=index)
                        except Exception as e:
                            self.gui.setStatus(
                                f"Scan warning: Can't plot Scan{len(self.datasets)-i}" \
                                    f" and dataframe {data_name} with ID {index+1}: {e}",
                                10000, False)
                        dataList2.append(data)

                    dataList2.reverse()
                    dataList.extend(dataList2)
            else:
                break

        dataList.reverse()
        return dataList

    def getLastDataset(self) -> Union[dict, None]:
        """ Returns the last created dataset """
        return self.datasets[-1] if len(self.datasets) > 0 else None

    def getLastSelectedDataset(self) -> Union[dict, None]:
        """ Returns the last selected dataset """
        index = self.gui.data_comboBox.currentIndex()
        if index != -1 and index < len(self.datasets):
            return self.datasets[index]
        return None

    def newDataset(self, config: dict):
        """ Creates and returns a new empty dataset """
        maximum = 0
        scanset = ScanSet()

        if self.save_temp:
            FOLDER_TEMP = os.environ['TEMP']  # This variable can be changed at autolab start-up
            folder_dataset_temp = tempfile.mkdtemp(dir=FOLDER_TEMP) # Creates a temporary directory for this dataset
            self.gui.configManager.export(
                os.path.join(folder_dataset_temp, 'config.conf'))
        else:
            folder_dataset_temp = str(random.random())

        for recipe_name, recipe in config.items():

            if recipe['active']:
                sub_folder = os.path.join(folder_dataset_temp, recipe_name)
                if self.save_temp: os.mkdir(sub_folder)

                dataset = Dataset(sub_folder, recipe_name,
                                  config, save_temp=self.save_temp)
                scanset[recipe_name] = dataset

                # bellow just to know maximum point
                nbpts = 1
                for parameter in recipe['parameter']:
                    if 'values' in parameter:
                        if has_eval(parameter['values']):
                            values = eval_safely(parameter['values'])
                            if isinstance(values, str):
                                nbpts *= 11  # OPTIMIZE: can't know length in this case without doing eval (should not do eval here because can imagine recipe_2 with param set at end of recipe_1)
                                self.gui.progressBar.setStyleSheet(
                                    "QProgressBar::chunk {background-color: orange;}")
                            else:
                                values = create_array(values)
                                nbpts *= len(values)
                        else: nbpts *= len(parameter['values'])
                    else: nbpts *= parameter['nbpts']

                maximum += nbpts

                # OBSOLETE: see if remove completely recipe in recipe
                list_recipe_nbpts_new = [[recipe, nbpts]]
                has_sub_recipe = True

                while has_sub_recipe:
                    has_sub_recipe = False
                    recipe_nbpts_list = list_recipe_nbpts_new

                    for recipe_nbpts in recipe_nbpts_list:
                        recipe_i, nbpts = recipe_nbpts

                        for step in recipe_i['recipe']:
                            if step['stepType'] == "recipe":
                                has_sub_recipe = True
                                other_recipe = config[step['element']]
                                other_nbpts = 1

                                for parameter in other_recipe['parameter']:
                                    if 'values' in parameter:
                                        other_nbpts *= len(parameter['values'])
                                    else:
                                        other_nbpts *= parameter['nbpts']
                                sub_nbpts = nbpts * other_nbpts
                                maximum += sub_nbpts
                                list_recipe_nbpts_new.append([other_recipe, sub_nbpts])

                        list_recipe_nbpts_new.remove(recipe_nbpts)

        self.datasets.append(scanset)
        self.gui.progressBar.setMaximum(maximum)

    def sync(self):
        """ This function sync the last dataset with the data available in the queue """
        # Empty the queue
        count = 0
        scanset = self.getLastDataset()
        lenQueue = self.queue.qsize()

        # Add scan data to dataset
        for _ in range(lenQueue):
            try: point = self.queue.get()  # point is collections.OrderedDict{0:recipe_name, 'parameter_name':parameter_value, 'step1_name':step1_value, 'step2_name':step2_value, ...}
            except: break

            recipe_name = list(point.values())[0]
            dataset = scanset[recipe_name]
            dataset.addPoint(point)
            count += 1

        # Upload the plot if new data available
        if count > 0:
            # Update progress bar
            progress = 0

            for dataset_name in self.gui.configManager.getRecipeActive():
                progress += len(scanset[dataset_name])

            self.gui.progressBar.setValue(progress)

            self.gui.save_pushButton.setEnabled(True)
            if len(self.datasets) != 1:
                self.gui.save_all_pushButton.setEnabled(True)

            # Update plot
            self.gui.figureManager.data_comboBoxClicked()

    def updateDisplayableResults(self):
        """ This function update the combobox in the GUI that displays the names of
        the results that can be plotted """
        data_name = self.gui.dataframe_comboBox.currentText()
        recipe_name = self.gui.scan_recipe_comboBox.currentText()
        scanset = self.getLastSelectedDataset()

        if scanset is None or recipe_name not in scanset: return None

        dataset = scanset[recipe_name]

        data = None
        if data_name == "Scan": data = dataset.data
        else:
            if dataset.data_arrays.get(data_name) is not None:
                for data in dataset.data_arrays[data_name]:  # used only to get columns name
                    if data is not None: break
            else: return None
            # if text or if image of type ndarray return
            if isinstance(data, str) or (
                    isinstance(data, np.ndarray) and not (
                        len(data.T.shape) == 1 or (
                            len(data.T.shape) != 0 and data.T.shape[0] == 2))):
                self.gui.variable_x_comboBox.clear()
                self.gui.variable_x2_comboBox.clear()
                self.gui.variable_y_comboBox.clear()
                return None

            try: data = data_to_dataframe(data)
            except AssertionError:  # if np.ndarray of string for example
                self.gui.variable_x_comboBox.clear()
                self.gui.variable_x2_comboBox.clear()
                self.gui.variable_y_comboBox.clear()
                return None

        result_names = []

        for result_name in data.columns:
            if result_name not in ['id']:
                try:
                    point = data.iloc[0][result_name]
                    if isinstance(point, pd.Series):
                        print('Warning: At least two variables have the same name.' \
                              f" Data acquisition is incorrect for {result_name}!",
                              file=sys.stderr)
                        float(point[0])
                    else:
                        float(point)
                    result_names.append(result_name)
                except:
                    pass

        variable_x_index = self.gui.variable_x_comboBox.currentIndex()
        variable_y_index = self.gui.variable_y_comboBox.currentIndex()

        items = [self.gui.variable_x_comboBox.itemText(i) for i in range(
            self.gui.variable_x_comboBox.count())]

        if result_names != items:  # only refresh if change labels, to avoid gui refresh that prevent user to click on combobox
            self.gui.variable_x_comboBox.clear()
            self.gui.variable_x2_comboBox.clear()
            self.gui.variable_x_comboBox.addItems(result_names)  # parameter first
            self.gui.variable_x2_comboBox.addItems(result_names)
            if result_names:
                name = result_names.pop(0)
                result_names.append(name)

            self.gui.variable_y_comboBox.clear()
            self.gui.variable_y_comboBox.addItems(result_names)  # first numerical measure first

            if data_name == "Scan":
                if variable_x_index != -1:
                    self.gui.variable_x_comboBox.setCurrentIndex(variable_x_index)
                    self.gui.variable_x2_comboBox.setCurrentIndex(variable_x_index)
                if variable_y_index != -1:
                    self.gui.variable_y_comboBox.setCurrentIndex(variable_y_index)
        return None


class Dataset():
    """ Collection of data from a recipe """
    def __init__(self, folder_dataset_temp: str, recipe_name: str, config: dict,
                 save_temp: bool = True):
        self._data_temp = []
        self.recipe_name = recipe_name
        self.folders = []
        self.data_arrays = {}
        self.folder_dataset_temp = folder_dataset_temp
        self.new = True
        self.save_temp = save_temp

        recipe = config[self.recipe_name]
        list_recipe = [recipe]
        list_recipe_new = [recipe]
        has_sub_recipe = True

        while has_sub_recipe:  # OBSOLETE
            has_sub_recipe = False
            recipe_list = list_recipe_new

            for recipe_i in recipe_list:
                for step in recipe_i['recipe']:
                    if step['stepType'] == "recipe":
                        has_sub_recipe = True
                        other_recipe = config[step['element']]
                        list_recipe_new.append(other_recipe)
                        list_recipe.append(other_recipe)

                list_recipe_new.remove(recipe_i)

        list_param = [recipe['parameter'] for recipe in list_recipe]
        self.list_param = sum(list_param, [])

        list_step = [recipe['recipe'] for recipe in list_recipe]
        self.list_step = sum(list_step, [])

        self.header = (["id"]
                       + [step['name'] for step in self.list_param]
                       + [step['name'] for step in self.list_step if (
                           step['stepType'] == 'measure'
                           and step['element'].type in [int, float, bool])]
                       )
        self.data = pd.DataFrame(columns=self.header)

    def getData(self, var_list: List[str], data_name: str = "Scan",
                dataID: int = 0, filter_condition: List[dict] = []) -> pd.DataFrame:
        """ This function returns a dataframe with two columns : the parameter value,
        and the requested result value """
        if data_name == "Scan":
            data = self.data
        else:
            data = self.data_arrays[data_name][dataID]

            if (data is not None
                    and not isinstance(data, str)
                    and (len(data.T.shape) == 1 or (
                        len(data.T.shape) != 0 and data.T.shape[0] == 2))):
                data = data_to_dataframe(data)
            else:  # Image
                return data

        # Add var for filtering
        for var_filter in filter_condition:
            if var_filter['enable']:
                if (var_filter['name'] not in var_list
                        and var_filter['name'] != ''
                        and var_filter['name'] is not None):
                    var_list.append(var_filter['name'])
                elif isinstance(var_filter['condition'], str):
                    for key in self.header:
                        if key in var_filter['condition']:
                            var_list.append(key)

        if any(map(lambda v: v in var_list, list(data.columns))):
            data = data.loc[:,~data.columns.duplicated()].copy()  # unique data column
            unique_var_list = list(dict.fromkeys(var_list))  # unique var_list
            # Filter data
            for var_filter in filter_condition:
                if var_filter['enable']:
                    if var_filter['name'] in data:
                        filter_cond = var_filter['condition']
                        filter_name = var_filter['name']
                        filter_value = var_filter['value']
                        mask = filter_cond(data[filter_name], filter_value)
                        data = data[mask]
                    elif isinstance(var_filter['condition'], str):
                        filter_cond = var_filter['condition']
                        if filter_cond:
                            try:
                                data = data.query(filter_cond)
                            except:
                                # If error, output empty dataframe
                                data = pd.DataFrame(columns=self.header)
                                break

            return data.loc[:,unique_var_list]

        return None

    def save(self, filename: str):
        """ This function saved the dataset in the provided path """
        dataset_folder = os.path.splitext(filename)[0]
        data_name = os.path.join(self.folder_dataset_temp, 'data.txt')

        if os.path.exists(data_name):
            shutil.copy(data_name, filename)
        else:
            self.data.to_csv(filename, index=False, header=self.header)

        if self.folders:
            if not os.path.exists(dataset_folder): os.mkdir(dataset_folder)
            for tmp_folder in self.folders:
                array_name = os.path.basename(tmp_folder)
                dest_folder = os.path.join(dataset_folder, array_name)

                if os.path.exists(tmp_folder):
                    try:
                        shutil.copytree(tmp_folder, dest_folder,
                                        dirs_exist_ok=True)  # python >=3.8 only
                    except:
                        if os.path.exists(dest_folder):
                            shutil.rmtree(dest_folder, ignore_errors=True)
                        shutil.copytree(tmp_folder, dest_folder)
                else:
                    # This is only executed if no temp folder is set
                    if not os.path.exists(dest_folder): os.mkdir(dest_folder)

                    if array_name in self.data_arrays:
                        list_data = self.data_arrays[array_name]  # data is list representing id 1,2

                        for i, value in enumerate(list_data):
                            path = os.path.join(dest_folder, f"{i+1}.txt")

                            if isinstance(value, (int, float, bool, str, tuple)):
                                with open(path, 'w') as f: f.write(str(value))
                            elif isinstance(value, bytes):
                                with open(path, 'wb') as f: f.write(value)
                            elif isinstance(value, np.ndarray):
                                df = pd.DataFrame(value)
                                df.to_csv(path, index=False, header=None)  # faster and handle better different dtype than np.savetxt
                            elif isinstance(value, pd.DataFrame):
                                value.to_csv(path, index=False)

    def addPoint(self, dataPoint: OrderedDict):
        """ This function add a data point (parameter value, and results) in the dataset """
        ID = len(self.data) + 1
        simpledata = OrderedDict()
        simpledata['id'] = ID

        for result_name, result in dataPoint.items():

            if result_name == 0: continue  # skip first result which is recipe_name

            elements = [step['element'] for step in (
                self.list_param+self.list_step) if step['name']==result_name]
            element = elements[0]
            # should always find exactly one element in list above

            # If the result is displayable (numerical), keep it in memory
            if element is None or element.type in [int, float, bool]:
                simpledata[result_name] = result
            else : # Else write it on a file, in a temp directory
                results_folder = os.path.join(self.folder_dataset_temp, result_name)

                if self.save_temp:
                    if not os.path.exists(results_folder): os.mkdir(results_folder)
                    result_path = os.path.join(results_folder, f'{ID}.txt')

                    if element is not None:
                        element.save(result_path, value=result)

                if results_folder not in self.folders:
                    self.folders.append(results_folder)

                if self.data_arrays.get(result_name) is None:
                    self.data_arrays[result_name] = []

                self.data_arrays[result_name].append(result)

        self._data_temp.append(simpledata)
        self.data = pd.DataFrame(self._data_temp, columns=self.header)

        if self.save_temp:
            if not os.path.exists(self.folder_dataset_temp):
                print(f'Warning: {self.folder_dataset_temp} has been created ' \
                      'but should have been created earlier. ' \
                      'Check that you have not lost any data',
                      file=sys.stderr)
                os.mkdir(self.folder_dataset_temp)
            if ID == 1:
                self.data.tail(1).to_csv(
                    os.path.join(self.folder_dataset_temp, 'data.txt'),
                    index=False, mode='a', header=self.header)
            else:
                self.data.tail(1).to_csv(
                    os.path.join(self.folder_dataset_temp, 'data.txt'),
                    index=False, mode='a', header=False)

    def __len__(self):
        """ Returns the number of data point of this dataset """
        return len(self.data)


class ScanSet(dict):
    """ Collection of data from a scan """
    # TODO: use this in scan plot
    display = True
    color = 'default'
    saved = False



# -*- coding: utf-8 -*-
"""
Created on Thu Sep  5 18:45:37 2024

@author: jonathan
"""

from typing import Dict


theme = {
    'dark': {
        'text_color': "#f0f0f0",
        'text_disabled_color': "#a0a0a0",
        'default': "#3c3c3c",
        'primary_color': "#3c3c3c",
        'secondary_color': "#2c2c2c",
        'tertiary_color': "#2e2e2e",
        'hover_color': "#555555",
        'pressed_color': "#666666",
        'selected_color': "#005777",
        'border_color': "#5c5c5c",
        'main_hover': "#005777",
        }
    }


def get_theme(theme_name: str) -> Dict[str, str]:
    return theme.get(theme_name)


def create_stylesheet(theme: Dict[str, str]) -> str:
    stylesheet = f"""
        QWidget {{
            background-color: {theme['default']};
            color: {theme['text_color']};
        }}
        QFrame {{
            background-color: {theme['primary_color']};
            border-radius: 8px;
        }}
        QFrame[frameShape='1'], QFrame[frameShape='2'], QFrame[frameShape='3'], QFrame[frameShape='6'] {{
            border: 1px solid {theme['border_color']};
        }}
        #line_V {{
            border-left: 1px solid {theme['border_color']};
        }}
        #line_H {{
            border-top: 1px solid {theme['border_color']};
        }}
        QPushButton {{
            background-color: {theme['secondary_color']};
            border: 1px solid {theme['border_color']};
            border-radius: 6px;
            padding: 4px;
            margin: 2px;
        }}
        QPushButton:hover {{
            background-color: {theme['hover_color']};
            border: 1px solid {theme['border_color']};
        }}
        QPushButton:pressed {{
            background-color: {theme['pressed_color']};
        }}
        QPushButton:disabled {{
            background-color: {theme['primary_color']};
            color: {theme['text_disabled_color']};
            border: 1px solid {theme['border_color']};
            border-radius: 6px;
        }}
        QCheckBox {{
            background-color: {theme['primary_color']};
            border-radius: 4px;
        }}
        QLineEdit, QTextEdit {{
            background-color: {theme['secondary_color']};
            border: 1px solid {theme['border_color']};
            border-radius: 4px;
        }}
        QLineEdit:disabled, QLineEdit[readOnly="true"] {{
            background-color: {theme['primary_color']};
            color: {theme['text_disabled_color']};
            border: 1px solid {theme['border_color']};
            border-radius: 4px;
        }}
        QComboBox {{
            background-color: {theme['secondary_color']};
            padding: 3px;
            border-radius: 5px;
        }}
        QTreeView {{
            alternate-background-color: {theme['secondary_color']};
            border-radius: 6px;
        }}
        QHeaderView::section {{
            background-color: {theme['secondary_color']};
            border: 1px solid {theme['primary_color']};
        }}
        QTabWidget::pane {{
            border: 1px solid {theme['border_color']};
            border-radius: 5px;
        }}
        QTabBar::tab {{
            background-color: {theme['tertiary_color']};
            border: 1px solid {theme['border_color']};
            padding: 5px;
            border-radius: 4px;
        }}
        QTabBar::tab:hover {{
            background-color: {theme['primary_color']};
            border: 1px solid {theme['hover_color']};
        }}
        QTabBar::tab:selected {{
            background-color: {theme['primary_color']};
            border: 1px solid {theme['hover_color']};
            margin-top: -2px;
            border-radius: 4px;
        }}
        QTabBar::tab:selected:hover {{
            background-color: {theme['primary_color']};
            border: 1px solid {theme['hover_color']};
            margin-top: -2px;
        }}
        QTabBar::tab:!selected {{
            margin-top: 2px;
        }}
        QMenuBar {{
            background-color: {theme['secondary_color']};
            border-radius: 5px;
        }}
        QMenuBar::item {{
            background-color: transparent;
            padding: 4px 10px;
            border-radius: 4px;
        }}
        QMenuBar::item:selected {{
            background-color: {theme['primary_color']};
            border: 1px solid {theme['pressed_color']};
        }}
        QMenu {{
            background-color: {theme['secondary_color']};
            border: 1px solid {theme['border_color']};
            border-radius: 6px;
        }}
        QMenu::item {{
            background-color: {theme['secondary_color']};
        }}
        QMenu::item:selected {{
            background-color: {theme['primary_color']};
        }}
        QMenu::item:disabled {{
            background-color: {theme['primary_color']};
            color: {theme['text_disabled_color']};
        }}
        QTreeWidget {{
            background-color: {theme['tertiary_color']};
            alternate-background-color: {theme['primary_color']};
        }}
        QTreeWidget::item {{
            border: none;
        }}
        QTreeWidget::item:hover {{
            background-color: {theme['main_hover']};
        }}
        QTreeWidget::item:selected {{
            background-color: {theme['selected_color']};
        }}
        QTreeWidget::item:selected:hover {{
            background-color: {theme['main_hover']};
        }}
        QScrollBar:vertical {{
            border: 1px solid {theme['primary_color']};
            background-color: {theme['tertiary_color']};
            width: 16px;
            border-radius: 6px;
        }}
        QScrollBar:horizontal {{
            border: 1px solid {theme['primary_color']};
            background-color: {theme['tertiary_color']};
            height: 16px;
            border-radius: 6px;
        }}
        QScrollBar::handle {{
            background-color: {theme['border_color']};
            border: 1px solid {theme['pressed_color']};
            border-radius: 6px;
        }}
        QScrollBar::handle:hover {{
            background-color: {theme['pressed_color']};
        }}
        QScrollBar::add-line, QScrollBar::sub-line {{
            background-color: {theme['primary_color']};
        }}
    """

    return stylesheet



# -*- coding: utf-8 -*-

import os
import shutil

from .paths import PATHS, DRIVER_LEGACY, DRIVER_SOURCES
from .repository import install_drivers


def process_all_changes():
    ''' Apply all changes '''
    rename_old_devices_config_file()
    move_driver()


def rename_old_devices_config_file():
    ''' Rename local_config.ini into devices_config.ini'''
    if (not os.path.exists(os.path.join(PATHS['user_folder'], 'devices_config.ini'))
            and os.path.exists(os.path.join(PATHS['user_folder'], 'local_config.ini'))):
        os.rename(os.path.join(PATHS['user_folder'], 'local_config.ini'),
                  os.path.join(PATHS['user_folder'], 'devices_config.ini'))


def move_driver():
    """ Move old driver directory to new one """
    if os.path.exists(os.path.join(PATHS['user_folder'])) and not os.path.exists(PATHS['drivers']):
        os.mkdir(PATHS['drivers'])
        print(f"The new driver directory has been created: {PATHS['drivers']}")

        # Inside os.path.exists(PATHS['drivers']) condition to avoid moving drivers from current repo everytime autolab is started
        if os.path.exists(DRIVER_LEGACY['official']):
            shutil.move(DRIVER_LEGACY['official'], PATHS['drivers'])
            os.rename(os.path.join(PATHS['drivers'], os.path.basename(DRIVER_LEGACY['official'])),
                      DRIVER_SOURCES['official'])
            print(f"Old official drivers directory has been moved from: {DRIVER_LEGACY['official']} to: {DRIVER_SOURCES['official']}")
            install_drivers()  # Ask if want to download official drivers

        if os.path.exists(DRIVER_LEGACY["local"]):
            shutil.move(DRIVER_LEGACY['local'], PATHS['drivers'])
            os.rename(os.path.join(PATHS['drivers'], os.path.basename(DRIVER_LEGACY['local'])),
                      DRIVER_SOURCES['local'])
            print(f"Old local drivers directory has been moved from: {DRIVER_LEGACY['local']} to: {DRIVER_SOURCES['local']}")



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 29 18:29:07 2019

@author: qchat
"""

from typing import Any, Union
import os

import pandas as pd
import numpy as np
from qtpy import QtCore, QtWidgets, QtGui

from ..icons import icons
from ..GUI_utilities import (MyLineEdit, MyInputDialog, MyQCheckBox, MyQComboBox,
                             qt_object_exists)
from ...paths import PATHS
from ...config import get_control_center_config
from ...variables import eval_variable, has_eval
from ...utilities import (SUPPORTED_EXTENSION, str_to_array, array_to_str,
                          dataframe_to_str, str_to_dataframe, create_array,
                          str_to_tuple)

# OPTIMIZE: Could merge treewidgets from control panel and plotter (or common version & subclass)

class TreeWidgetItemModule(QtWidgets.QTreeWidgetItem):
    """ This class represents a module in an item of the tree """

    def __init__(self, itemParent, name, nickname, gui):

        self.name = name
        self.module = None
        self.loaded = False
        self.gui = gui
        self.is_not_submodule = isinstance(gui.tree, type(itemParent))
        self.nickname = nickname

        super().__init__(itemParent, [nickname, 'Module'])

        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

    def load(self, module):
        """ This function loads the entire module (submodules, variables, actions) """
        self.module = module

        # Submodules
        subModuleNames = self.module.list_modules()
        for subModuleName in subModuleNames:
            subModule = getattr(self.module, subModuleName)
            item = TreeWidgetItemModule(self, subModuleName, subModuleName, self.gui)
            item.load(subModule)

        # Variables
        varNames = self.module.list_variables()
        for varName in varNames:
            variable = getattr(self.module,varName)
            TreeWidgetItemVariable(self, variable, self.gui)

        # Actions
        actNames = self.module.list_actions()
        for actName in actNames:
            action = getattr(self.module,actName)
            TreeWidgetItemAction(self, action, self.gui)

        # Change loaded status
        self.loaded = True

        # Tooltip
        if self.module._help is not None: self.setToolTip(0, self.module._help)

    def menu(self, position):
        """ This function provides the menu when the user right click on an item """
        if self.is_not_submodule and self.loaded:
            menu = QtWidgets.QMenu()
            disconnectDevice = menu.addAction(f"Disconnect {self.nickname}")

            choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))

            if choice == disconnectDevice:
                device = self.gui.active_plugin_dict[self.nickname]
                try: device.instance.close()  # not device close because device.close will remove device from DEVICES list
                except: pass
                self.gui.active_plugin_dict.pop(self.nickname)

                for i in range(self.childCount()):
                    self.removeChild(self.child(0))
                self.loaded = False

                if not self.gui.active_plugin_dict:
                    self.gui._stop_timerQueue = True


class TreeWidgetItemAction(QtWidgets.QTreeWidgetItem):
    """ This class represents an action in an item of the tree """

    def __init__(self, itemParent, action, gui):

        displayName = f'{action.name}'
        if action.unit is not None:
            displayName += f' ({action.unit})'

        super().__init__(itemParent, [displayName, 'Action'])
        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

        self.gui = gui
        self.action = action

        # Import Autolab config
        control_center_config = get_control_center_config()
        self.precision = int(float(control_center_config['precision']))

        if self.action.has_parameter:
            if self.action.type in [int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
                self.executable = True
                self.has_value = True
            else:
                self.executable = False
                self.has_value = False
        else:
            self.executable = True
            self.has_value = False

        # Main - Column 2 : actionread button
        if self.executable:
            self.execButton = QtWidgets.QPushButton()
            self.execButton.setText("Execute")
            self.execButton.clicked.connect(self.execute)
            self.gui.tree.setItemWidget(self, 2, self.execButton)

        # Main - Column 3 : QlineEdit if the action has a parameter
        if self.has_value:
            if self.action.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
                self.valueWidget = MyLineEdit()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
                self.gui.tree.setItemWidget(self, 3, self.valueWidget)
                self.valueWidget.returnPressed.connect(self.execute)
                self.valueWidget.textEdited.connect(self.valueEdited)

            ## QCheckbox for boolean variables
            elif self.action.type in [bool]:
                self.valueWidget = MyQCheckBox(self)
                hbox = QtWidgets.QHBoxLayout()
                hbox.addWidget(self.valueWidget)
                hbox.setAlignment(QtCore.Qt.AlignCenter)
                hbox.setSpacing(0)
                hbox.setContentsMargins(0,0,0,0)
                widget = QtWidgets.QWidget()
                widget.setLayout(hbox)

                self.gui.tree.setItemWidget(self, 3, widget)

            ## Combobox for tuples: Tuple[List[str], int]
            elif self.action.type in [tuple]:
                self.valueWidget = MyQComboBox()
                self.valueWidget.wheel = False  # prevent changing value by mistake
                self.valueWidget.key = False
                self.valueWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
                self.valueWidget.customContextMenuRequested.connect(self.openInputDialog)

                self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        # Main - column 4 : indicator (status of the actual value : known or not known)
        self.indicator = QtWidgets.QLabel()
        self.gui.tree.setItemWidget(self, 4, self.indicator)

        # Tooltip
        if self.action._help is None: tooltip = 'No help available for this action'
        else: tooltip = self.action._help
        self.setToolTip(0, tooltip)

        self.writeSignal = WriteSignal()
        self.writeSignal.writed.connect(self.valueWrited)
        self.action._write_signal = self.writeSignal

    def openInputDialog(self, position: QtCore.QPoint):
        """ Only used for tuple """
        menu = QtWidgets.QMenu()
        modifyTuple = menu.addAction("Modify tuple")
        modifyTuple.setIcon(icons['tuple'])

        choice = menu.exec_(self.valueWidget.mapToGlobal(position))

        if choice == modifyTuple:
            main_dialog = MyInputDialog(self.gui, self.action.address())
            main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
            if self.action.type in [tuple]:
                main_dialog.setTextValue(str(self.action.value))
            main_dialog.show()

            if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                response = main_dialog.textValue()
            else:
                response = ''

            if qt_object_exists(main_dialog): main_dialog.deleteLater()

            if response != '':
                try:
                    if has_eval(response):
                        response = eval_variable(response)
                    if self.action.type in [tuple]:
                        response = str_to_tuple(str(response))
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.action.address()}: {e}", 10000, False)
                    return None

                self.action.value = response
                self.valueWrited(response)
                self.valueEdited()

    def writeGui(self, value):
        """ This function displays a new value in the GUI """
        if hasattr(self, 'valueWidget') and qt_object_exists(self.valueWidget):  # avoid crash if device closed and try to write gui (if close device before reading finished)
            # Update value
            if self.action.type in [int, float]:
                self.valueWidget.setText(f'{value:.{self.precision}g}') # default is .6g
            elif self.action.type in [str]:
                self.valueWidget.setText(value)
            elif self.action.type in [bytes]:
                self.valueWidget.setText(value.decode())
            elif self.action.type in [bool]:
                self.valueWidget.setChecked(value)
            elif self.action.type in [tuple]:
                items = [self.valueWidget.itemText(i)
                         for i in range(self.valueWidget.count())]
                if value[0] != items:
                    self.valueWidget.clear()
                    self.valueWidget.addItems(value[0])
                self.valueWidget.setCurrentIndex(value[1])
            elif self.action.type in [np.ndarray]:
                self.valueWidget.setText(array_to_str(value))
            elif self.action.type in [pd.DataFrame]:
                self.valueWidget.setText(dataframe_to_str(value))
            else:
                self.valueWidget.setText(value)

    def readGui(self) -> Any:
        """ This function returns the value in good format of the value in the GUI """
        if self.action.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            value = self.valueWidget.text()

            if value == '':
                if self.action.unit in ('open-file', 'save-file', 'filename'):
                    if self.action.unit == "filename":  # TODO: LEGACY (to remove later)
                        self.gui.setStatus("Using 'filename' as unit is depreciated in favor of 'open-file' and 'save-file'" \
                                           f"\nUpdate driver '{self.action.address().split('.')[0]}' to remove this warning",
                                           10000, False)
                        self.action.unit = "open-file"

                    if self.action.unit == "open-file":
                        filename, _ = QtWidgets.QFileDialog.getOpenFileName(
                            self.gui, caption=f"Open file - {self.action.address()}",
                            directory=PATHS['last_folder'],
                            filter=SUPPORTED_EXTENSION)
                    elif self.action.unit == "save-file":
                        filename, _ = QtWidgets.QFileDialog.getSaveFileName(
                            self.gui, caption=f"Save file - {self.action.address()}",
                            directory=PATHS['last_folder'],
                            filter=SUPPORTED_EXTENSION)

                    if filename != '':
                        path = os.path.dirname(filename)
                        PATHS['last_folder'] = path
                        return filename
                    else:
                        self.gui.setStatus(
                            f"Action {self.action.address()} cancel filename selection",
                            10000)
                elif self.action.unit == "user-input":
                    main_dialog = MyInputDialog(self.gui, self.action.address())
                    main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
                    main_dialog.show()

                    if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                        response = main_dialog.textValue()
                    else:
                        response = ''

                    if qt_object_exists(main_dialog): main_dialog.deleteLater()

                    if response != '':
                        return response
                    else:
                        self.gui.setStatus(
                            f"Action {self.action.address()} cancel user input",
                            10000)
                else:
                    self.gui.setStatus(
                        f"Action {self.action.address()} requires a value for its parameter",
                        10000, False)
            else:
                try:
                    value = eval_variable(value)
                    if self.action.type in [int]:
                        value = int(float(value))
                    if self.action.type in [bytes]:
                        value = value.encode()
                    elif self.action.type in [np.ndarray]:
                        value = str_to_array(value) if isinstance(
                            value, str) else create_array(value)
                    elif self.action.type in [pd.DataFrame]:
                        if isinstance(value, str):
                            value = str_to_dataframe(value)
                    else:
                        value = self.action.type(value)
                    return value
                except Exception as e:
                    self.gui.setStatus(
                        f"Action {self.action.address()}: {e}",
                        10000, False)
        elif self.action.type in [bool]:
            value = self.valueWidget.isChecked()
            return value
        elif self.action.type in [tuple]:
            items = [self.valueWidget.itemText(i)
                     for i in range(self.valueWidget.count())]
            value = (items, self.valueWidget.currentIndex())
            return value

    def setValueKnownState(self, state: Union[bool, float]):
        """ Turn the color of the indicator depending of the known state of the value """
        if state == 0.5:
            self.indicator.setStyleSheet("background-color:#FFFF00")  # yellow
            self.indicator.setToolTip('Value written but not read')
        elif state:
            self.indicator.setStyleSheet("background-color:#70db70")  # green
            self.indicator.setToolTip('Value read')
        else:
            self.indicator.setStyleSheet("background-color:#ff8c1a")  # orange
            self.indicator.setToolTip('Value not up-to-date')

    def execute(self):
        """ Start a new thread to execute the associated action """
        if not self.isDisabled():
            if self.has_value:
                value = self.readGui()
                if value is not None:
                    self.gui.threadManager.start(self, 'execute', value=value)
            else:
                self.gui.threadManager.start(self, 'execute')

    def valueEdited(self):
        """ Change indicator state when editing action parameter """
        self.setValueKnownState(False)

    def valueWrited(self, value: Any):
        """ Called when action parameter written """
        try:
            if self.has_value:
                self.writeGui(value)
            self.setValueKnownState(0.5)
        except Exception as e:
            self.gui.setStatus(f"SHOULD NOT RAISE ERROR: {e}", 10000, False)


class TreeWidgetItemVariable(QtWidgets.QTreeWidgetItem):
    """ This class represents a variable in an item of the tree """

    def __init__(self, itemParent, variable, gui):

        displayName = f'{variable.name}'
        if variable.unit is not None:
            displayName += f' ({variable.unit})'

        super().__init__(itemParent, [displayName, 'Variable'])
        self.setTextAlignment(1, QtCore.Qt.AlignHCenter)

        self.gui = gui
        self.variable = variable

        # Import Autolab config
        control_center_config = get_control_center_config()
        self.precision = int(float(control_center_config['precision']))

        # Signal creation and associations in autolab devices instances
        self.readSignal = ReadSignal()
        self.readSignal.read.connect(self.writeGui)
        self.variable._read_signal = self.readSignal
        self.writeSignal = WriteSignal()
        self.writeSignal.writed.connect(self.valueWrited)
        self.variable._write_signal = self.writeSignal

        # Main - Column 2 : Creation of a READ button if the variable is readable
        if self.variable.readable and self.variable.type in [
                int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
            self.readButton = QtWidgets.QPushButton()
            self.readButton.setText("Read")
            self.readButton.clicked.connect(self.read)

            if not self.variable.writable and self.variable.type in [
                    np.ndarray, pd.DataFrame]:
                self.readButtonCheck = QtWidgets.QCheckBox()
                self.readButtonCheck.stateChanged.connect(
                    self.readButtonCheckEdited)
                self.readButtonCheck.setToolTip(
                    'Toggle reading in text, ' \
                    'careful can truncate data and impact performance')
                self.readButtonCheck.setMaximumWidth(15)

                frameReadButton = QtWidgets.QFrame()
                hbox = QtWidgets.QHBoxLayout(frameReadButton)
                hbox.setSpacing(0)
                hbox.setContentsMargins(0,0,0,0)
                hbox.addWidget(self.readButtonCheck)
                hbox.addWidget(self.readButton)
                self.gui.tree.setItemWidget(self, 2, frameReadButton)
            else:
                self.gui.tree.setItemWidget(self, 2, self.readButton)

        # Main - column 3 : Creation of a VALUE widget, depending on the type

        ## QLineEdit or QLabel
        if self.variable.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            if self.variable.writable:
                self.valueWidget = MyLineEdit()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
                self.valueWidget.returnPressed.connect(self.write)
                self.valueWidget.textEdited.connect(self.valueEdited)
                self.valueWidget.setMaxLength(10000000)  # default is 32767, not enought for array and dataframe
                # self.valueWidget.setPlaceholderText(self.variable._help)  # Could be nice but take too much place. Maybe add it as option
            elif self.variable.readable:
                self.valueWidget = QtWidgets.QLineEdit()
                self.valueWidget.setMaxLength(10000000)
                self.valueWidget.setReadOnly(True)
                palette = self.valueWidget.palette()
                palette.setColor(QtGui.QPalette.Base,
                                 palette.color(QtGui.QPalette.Base).darker(107))
                self.valueWidget.setPalette(palette)
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)
            else:
                self.valueWidget = QtWidgets.QLabel()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)

            self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        ## QCheckbox for boolean variables
        elif self.variable.type in [bool]:
            self.valueWidget = MyQCheckBox(self)
            hbox = QtWidgets.QHBoxLayout()
            hbox.addWidget(self.valueWidget)
            hbox.setAlignment(QtCore.Qt.AlignCenter)
            hbox.setSpacing(0)
            hbox.setContentsMargins(0,0,0,0)
            widget = QtWidgets.QWidget()
            widget.setLayout(hbox)
            if not self.variable.writable:  # Disable interaction is not writable
                self.valueWidget.setEnabled(False)

            self.gui.tree.setItemWidget(self, 3, widget)

        ## Combobox for tuples: Tuple[List[str], int]
        elif self.variable.type in [tuple]:
            if self.variable.writable:
                self.valueWidget = MyQComboBox()
                self.valueWidget.wheel = False  # prevent changing value by mistake
                self.valueWidget.key = False
                self.valueWidget.activated.connect(self.write)
                self.valueWidget.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
                self.valueWidget.customContextMenuRequested.connect(self.openInputDialog)
            elif self.variable.readable:
                self.valueWidget = MyQComboBox()
                self.valueWidget.readonly = True
            else:
                self.valueWidget = QtWidgets.QLabel()
                self.valueWidget.setAlignment(QtCore.Qt.AlignCenter)

            self.gui.tree.setItemWidget(self, 3, self.valueWidget)

        # Main - column 4 : indicator (status of the actual value : known or not known)
        self.indicator = QtWidgets.QLabel()
        self.gui.tree.setItemWidget(self, 4, self.indicator)

        # Tooltip
        if self.variable._help is None: tooltip = 'No help available for this variable'
        else: tooltip = self.variable._help
        if hasattr(self.variable, "type"):
            variable_type = str(self.variable.type).split("'")[1]
            tooltip += f" ({variable_type})"
        self.setToolTip(0, tooltip)

    def openInputDialog(self, position: QtCore.QPoint):
        """ Only used for tuple """
        menu = QtWidgets.QMenu()
        modifyTuple = menu.addAction("Modify tuple")
        modifyTuple.setIcon(icons['tuple'])

        choice = menu.exec_(self.valueWidget.mapToGlobal(position))

        if choice == modifyTuple:
            main_dialog = MyInputDialog(self.gui, self.variable.address())
            main_dialog.setWindowModality(QtCore.Qt.ApplicationModal)
            if self.variable.type in [tuple]:
                main_dialog.setTextValue(str(self.variable.value))
            main_dialog.show()

            if main_dialog.exec_() == QtWidgets.QInputDialog.Accepted:
                response = main_dialog.textValue()
            else:
                response = ''

            if qt_object_exists(main_dialog): main_dialog.deleteLater()

            if response != '':
                try:
                    if has_eval(response):
                        response = eval_variable(response)
                    if self.variable.type in [tuple]:
                        response = str_to_tuple(str(response))
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.variable.address()}: {e}", 10000, False)
                    return None

                self.variable(response)

                if self.variable.readable:
                    self.variable()

    def writeGui(self, value):
        """ This function displays a new value in the GUI """
        if hasattr(self, 'valueWidget') and qt_object_exists(self.valueWidget):  # avoid crash if device closed and try to write gui (if close device before reading finished)
            # Update value
            if self.variable.numerical:
                self.valueWidget.setText(f'{value:.{self.precision}g}') # default is .6g
            elif self.variable.type in [str]:
                self.valueWidget.setText(value)
            elif self.variable.type in [bytes]:
                self.valueWidget.setText(value.decode())
            elif self.variable.type in [bool]:
                self.valueWidget.setChecked(value)
            elif self.variable.type in [tuple]:
                items = [self.valueWidget.itemText(i)
                         for i in range(self.valueWidget.count())]
                if value[0] != items:
                    self.valueWidget.clear()
                    self.valueWidget.addItems(value[0])
                self.valueWidget.setCurrentIndex(value[1])
            elif self.variable.type in [np.ndarray, pd.DataFrame]:
                if self.variable.writable or self.readButtonCheck.isChecked():
                    if self.variable.type in [np.ndarray]:
                        self.valueWidget.setText(array_to_str(value))
                    if self.variable.type in [pd.DataFrame]:
                        self.valueWidget.setText(dataframe_to_str(value))
                # else:
                #     self.valueWidget.setText('')

            # Change indicator light to green
            if self.variable.type in [
                    int, float, bool, str, bytes, tuple, np.ndarray, pd.DataFrame]:
                self.setValueKnownState(True)

    def readGui(self):
        """ This function returns the value in good format of the value in the GUI """
        if self.variable.type in [int, float, str, bytes, np.ndarray, pd.DataFrame]:
            value = self.valueWidget.text()
            if value == '':
                self.gui.setStatus(
                    f"Variable {self.variable.address()} requires a value to be set",
                    10000, False)
            else:
                try:
                    value = eval_variable(value)
                    if self.variable.type in [int]:
                        value = int(float(value))
                    if self.variable.type in [bytes]:
                        value = value.encode()
                    elif self.variable.type in [np.ndarray]:
                        if isinstance(value, str): value = str_to_array(value)
                        else: value = create_array(value)
                    elif self.variable.type in [pd.DataFrame]:
                        if isinstance(value, str): value = str_to_dataframe(value)
                    else:
                        value = self.variable.type(value)
                    return value
                except Exception as e:
                    self.gui.setStatus(
                        f"Variable {self.variable.address()}: {e}",
                        10000, False)

        elif self.variable.type in [bool]:
            value = self.valueWidget.isChecked()
            return value
        elif self.variable.type in [tuple]:
            items = [self.valueWidget.itemText(i)
                     for i in range(self.valueWidget.count())]
            value = (items, self.valueWidget.currentIndex())
            return value

    def setValueKnownState(self, state: Union[bool, float]):
        """ Turn the color of the indicator depending of the known state of the value """
        if state == 0.5:
            self.indicator.setStyleSheet("background-color:#FFFF00")  # yellow
            self.indicator.setToolTip('Value written but not read')
        elif state:
            self.indicator.setStyleSheet("background-color:#70db70")  # green
            self.indicator.setToolTip('Value read')
        else:
            self.indicator.setStyleSheet("background-color:#ff8c1a")  # orange
            self.indicator.setToolTip('Value not up-to-date')

    def read(self):
        """ Start a new thread to READ the associated variable """
        if not self.isDisabled():
            self.setValueKnownState(False)
            self.gui.threadManager.start(self, 'read')

    def write(self):
        """ Start a new thread to WRITE the associated variable """
        if not self.isDisabled():
            value = self.readGui()
            if value is not None:
                self.gui.threadManager.start(self, 'write', value=value)

    def valueWrited(self, value: Any):
        """ Function call when the value displayed in not sure anymore.
            The value has been modified either in the GUI (but not sent)
            or by command line.
            If variable not readable, write the value sent to the GUI """
        # BUG: I got an error when changing emit_write to set value, need to reproduce it
        try:
            self.writeGui(value)
            self.setValueKnownState(0.5)
        except Exception as e:
            self.gui.setStatus(f"SHOULD NOT RAISE ERROR: {e}", 10000, False)

    def valueEdited(self):
        """ Function call when the value displayed in not sure anymore.
            The value has been modified either in the GUI (but not sent)
            or by command line """
        self.setValueKnownState(False)

    def readButtonCheckEdited(self):
        state = bool(self.readButtonCheck.isChecked())
        self.readButton.setEnabled(state)

    def menu(self, position):
        """ This function provides the menu when the user right click on an item """
        if not self.isDisabled():
            menu = QtWidgets.QMenu()

            saveAction = menu.addAction("Read and save as...")
            saveAction.setEnabled(self.variable.readable)

            choice = menu.exec_(self.gui.tree.viewport().mapToGlobal(position))

            if choice == saveAction:
                self.saveValue()

    def saveValue(self):
        filename = QtWidgets.QFileDialog.getSaveFileName(
            self.gui, f"Save {self.variable.name} value", os.path.join(
                PATHS['last_folder'],f'{self.variable.address()}.txt'),
            filter=SUPPORTED_EXTENSION)[0]

        path = os.path.dirname(filename)
        if path != '':
            PATHS['last_folder'] = path
            try:
                self.gui.setStatus(
                    f"Saving value of {self.variable.name}...", 5000)
                self.variable.save(filename)
                self.gui.setStatus(
                    f"Value of {self.variable.name} successfully read and save at {filename}",
                    5000)
            except Exception as e:
                self.gui.setStatus(f"An error occured: {e}", 10000, False)


# Signals can be emitted only from QObjects
# These class provides convenient ways to use signals
class ReadSignal(QtCore.QObject):
    read = QtCore.Signal(object)
    def emit_read(self, value):
        self.read.emit(value)

class WriteSignal(QtCore.QObject):
    writed = QtCore.Signal(object)
    def emit_write(self, value):
        self.writed.emit(value)



# -*- coding: utf-8 -*-
"""
Created on Fri Sep 20 17:11:57 2019

@author: qchat
"""

import webbrowser
import socket
import os
import inspect

from .utilities import open_file

project_url = 'https://github.com/autolab-project/autolab'
drivers_url = 'https://github.com/autolab-project/autolab-drivers'
doc_url = 'https://autolab.readthedocs.io'


def report():
    """ Open the github link to open an issue: https://github.com/autolab-project/autolab/issues """
    webbrowser.open(project_url+'/issues')


def doc(online: bool = "default"):
    """ By default try to open the online doc and if no internet connection, open the local pdf documentation.
    Can open online or offline documentation by using True or False."""

    if online == "default":
        if has_internet(False): webbrowser.open(doc_url)
        else:
            print("No internet connection found. Open local pdf documentation instead")
            doc_offline()
    elif online: webbrowser.open(doc_url)
    elif not online: doc_offline()


def doc_offline():
    dirname = os.path.dirname(os.path.abspath(inspect.stack()[0][1]))
    filename = os.path.join(dirname, "../autolab.pdf")
    if os.path.exists(filename): open_file(filename)
    else: print("No local pdf documentation found at {filename}")


def has_internet(_print=True) -> bool:
    """ https://stackoverflow.com/questions/20913411/test-if-an-internet-connection-is-present-in-python#20913928 """
    try:
        # see if we can resolve the host name -- tells us if there is
        # a DNS listening
        host = socket.gethostbyname("one.one.one.one")
        # connect to the host -- tells us if the host is actually reachable
        s = socket.create_connection((host, 80), 2)
        s.close()
        return True
    except Exception: pass # we ignore any errors, returning False

    if _print: print("No internet connection found")
    return False



# -*- coding: utf-8 -*-
"""
Created on Sat Jul 20 17:30:03 2019

@author: qchat
"""

from setuptools import setup,find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

with open("./autolab/version.txt", "r") as fh:
    version = fh.read().strip()

# Use python setup.py install to test if package is correct

setup(
    name = 'autolab',
    version = version,  # Ideally should be same as your GitHub release tag varsion
    author = 'Quentin Chateiller & Bruno Garbin',
    author_email = 'autolab-project@googlegroups.com',
    maintainer = 'Jonathan Peltier & Mathieu Jeannin',
    maintainer_email = 'autolab-project@googlegroups.com',
    license = "GPL-3.0 license",
    description = 'Python package for scientific experiments interfacing and automation',
    long_description = long_description,
    long_description_content_type="text/markdown",
    url = 'https://github.com/autolab-project/autolab',
    packages=find_packages(),
    classifiers=["Programming Language :: Python :: 3",
                 "Programming Language :: Python :: 3 :: Only",
				"Programming Language :: Python :: 3.6",
				"Programming Language :: Python :: 3.7",
				"Programming Language :: Python :: 3.8",
				"Programming Language :: Python :: 3.9",
				"Programming Language :: Python :: 3.10",
				"Programming Language :: Python :: 3.11",
				"Programming Language :: Python :: 3.12",
                 "Operating System :: OS Independent"],
    install_requires=[
            'numpy>=1.16',
            'pandas>=0.24',
            'pyvisa>=1.10',
            'python-vxi11>=0.9',
            'qtpy',
            'pyqtgraph',
            'requests',
            'tqdm',
            'comtypes',
            ],
    entry_points={'console_scripts': ['autolab = autolab:_main']},
	python_requires='>=3.6',
    include_package_data=True,
    package_data={'': ['*.ini','*.txt','*.ui']},# If any package contains *.ini files, include them:
    keywords = ['scanning','interface','automation','scientific','laboratory','devices','experiments','measures','interface','gui','scan']
)



# -*- coding: utf-8 -*-

import socket
import pickle
import threading
import datetime as dt
from functools import partial

from .config import get_server_config
from .devices import get_devices_status, get_device


class Driver_SOCKET():

    prefix = b'<AUTOLAB_START>'
    suffix = b'<AUTOLAB_END>'

    def read(self, length=4096):
        ''' Read pickled object from autolab master and return python object '''
        # First read
        msg  = self.socket.recv(length)
        assert msg != b'', 'Connection closed by remote host'
        assert msg.startswith(self.prefix), 'Autolab communication structure not found in reply'

        # Continue reading up to suffix
        while not msg.endswith(self.suffix):
            msg += self.socket.recv(length)

        # Clean msg (remove prefix and suffix) and unpicke it
        msg = msg.lstrip(self.prefix).rstrip(self.suffix)
        obj = pickle.loads(msg)

        return obj

    def write(self, object):
        ''' Send pickled object to autolab master '''
        msg = self.prefix + pickle.dumps(object) + self.suffix
        self.socket.send(msg)


class ClientThread(threading.Thread, Driver_SOCKET):

    def __init__(self, client_socket, server):
        super().__init__()
        self.socket = client_socket
        self.server = server
        self.stop_flag = threading.Event()
        self.hostname = None

    def run(self):

        # Handshaking
        if self.handshake():
            # Start listening client commands
            self.listen()

        # Close client socket
        self.close()

    def handshake(self) -> bool:

        ''' Check that incoming connection comes from another Autolab program,
            and that no thread is already controlling autolab '''

        self.socket.settimeout(2)

        try :
            # Read first command from client
            handshake_str = self.read()

            # Check that first client command is 'AUTOLAB?'
            if handshake_str.startswith('AUTOLAB?'):

                self.hostname = handshake_str.split('=')[1]

                # There is no main thread currently running, accepting communication
                if self.server.active_connection_thread is None:
                    self.server.active_connection_thread = self
                    self.server.log(f'Host "{self.hostname}" connected')
                    self.write('YES')
                    result = True

                # Another client is controlling autolab, reply that the server is busy, refusing communication
                else :
                    self.server.log(f'Host "{self.hostname}" trying to connect but a connection is already established from host "{self.server.active_connection_thread.hostname}"')
                    self.write(f'A connection to the Autolab server is already established from host "{self.server.active_connection_thread.hostname}".')
                    result = False

            # The client did not ask the right first command, refusing communication
            else: result = False

        except Exception as e:
            print(e)
            result = False

        self.socket.settimeout(None)

        return result

    def listen(self):
        ''' Listen and answer client commands '''
        while not self.stop_flag.is_set():
            try:
                command = self.read()
                self.process_command(command)
            except:
                self.stop_flag.set()

    def process_command(self, command):
        ''' Process given client command '''
        if isinstance(command, str):
            if command == 'CLOSE_CONNECTION':
                self.stop_flag.set()
            elif command == 'DEVICES_STATUS?':
                return self.write(get_devices_status())
        else:
            if command['command'] == 'get_device_model':
                device_name = command['device_name']
                structure = get_device(device_name).get_structure()
                self.write(structure)
            elif command['command'] == 'request':
                get_device(device_name).get_by_adress(command['element_adress'])
                # element_address --> my_yenista::submodule::wavelength
                wavelength()

#command_dict = {'command':'get_driver_model','name':self.name}

    def close(self):

        # In case the close call come from outside of the threading
        self.socket.close()

        # If this thread is the main client thread, remove declaration
        if self.server.active_connection_thread == self:
            self.server.log(f'Host "{self.hostname}" disconnected')
            self.server.active_connection_thread = None

    def request_close(self):

        self.stop_flag.set()
        self.socket.shutdown(socket.SHUT_RDWR)
        self.join()


class Server():

    def __init__(self, port=None):

        self.client_threads = []
        self.active_connection_thread = None

        # Load server config in autolab_config.ini
        server_config = get_server_config()
        if not port: port = int(server_config['port'])
        self.port = port

        # Start the server
        self.start()

        # Start listening
        try: self.listen()
        except: print('You excited the server (TO CHANGE)')
        self.close()


    def start(self):
        ''' Start the server '''
        self.main_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.main_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.main_socket.bind(('', self.port))
        self.main_socket.listen(0)
        self.log(f'Autolab server running, waiting for incoming connections on port {self.port}')


    def listen(self):

        while True:

            # Wait incoming connection
            client_socket, _ = self.main_socket.accept()

            # Clean terminated threads
            self.clean_client_threads()

            # Start new client thread
            client_thread = ClientThread(client_socket, self)
            client_thread.start()
            self.client_threads.append(client_thread)


    def clean_client_threads(self):
        ''' Remove finshed client threads from the list '''
        self.client_threads = [t for t in self.client_threads if t.is_alive()]

    def close_client_threads(self):
        ''' Close any existing client thread '''
        self.clean_client_threads()

        for thread in self.client_threads:
            thread.request_close()

        self.client_threads = []


    def log(self, log):
        ''' Display a log on the server '''
        timestamp = dt.datetime.now().isoformat()
        print(f'{timestamp}: {log}')


    def close(self):
        ''' Close the server and client threads'''
        self.close_client_threads()
        self.main_socket.shutdown(socket.SHUT_RDWR)
        self.main_socket.close()


class Driver_REMOTE(Driver_SOCKET):

    def __init__(self, address='192.168.1.1', port=4001):
        from functools import partial

        self.address = address
        self.port = int(port)

        # Connection au serveur Autolab
        self.connect()

        # Handshaking
        self.handshake()

        # Retourne la liste des devices
        self.devices_status = self.get_devices_status()
        print(self.devices_status)


    def connect(self):
        ''' Initialize connection with autolab server '''
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.socket.settimeout(2)
        self.socket.connect((self.address, self.port))
        self.socket.settimeout(None)


    def handshake(self):
        ''' Check that distant partner is an Autolab server '''
        self.write(f'AUTOLAB?HOSTNAME={socket.gethostname()}')
        try:
            answer = self.read()
            assert answer == 'YES', answer
        except Exception as e:
            raise ValueError(f'Impossible to join autolab server at {self.address}:{self.port} \n {e}')

    def disconnect(self):
        ''' Close autolab server connection '''
        self.socket.write('CLOSE_CONNECTION')
        self.socket.shutdown(socket.SHUT_RDWR)
        self.socket.close()

    def get_devices_status(self): # Déjà instantié ou non
        self.write('DEVICES_STATUS?')
        return self.read()

    def get_driver_model(self):
        model = []
        for dev_name in self.devices_status.keys():
            a = FakeDriver(self, dev_name, status)
            model.append({'element': 'module', 'name': dev_name, 'object': a})
            a.get_driver_model()
            partial(devices.Device, dev_name)
        return model

    def get_device(self, dev_name):
        pass

    def get_device_structure(self):
        pass
            # Pour un device de slave donné, retourne sa structure de (sous-devices, modules, action variable) à l’instant t
            # Device non instantité : ne renvoie rien
            # Device instantié : renvoie structure en (modules, action, variable)


class FakeDriver():

    def __init__(self, driver_remote, name, status):

        self.driver_remote = driver_remote
        self.name = name
        self.status = status

    def get_driver_model(self):
        model = []
        if status is False:
            return model
        else:
            command_dict = {'command': 'get_device_structure', 'name': self.name}
            remote_device_structure = self.driver_remote.write(command_dict)

            for key in remote_device_structure.keys():
                if remote_device_structure[key] == 'variable':
                    arg = {'command': 'request', address: 'yenista::wavelength'}
                    #model.append({'element':remote_device_structure[key], 'name':key, 'read': partial(self.interact,'read',address) .., 'write'})
                elif remote_device_structure[key] == 'action':
                    pass
                elif remote_device_structure[key] == 'module':
                    pass

            return remote_driver_model

    def interact(interaction_type, address):
        pass


#{'slot1': {'power': 'variable', 'wavelength': 'variable'},
# 'amplitude': 'variable',
# 'something': 'action'}

#def get_driver_model(self):

 #    model = []
#    model.append({'element':'variable','name':'amplitude','write':self.amplitude,'read':self.get_amplitude,'unit':'V','type':float,'help':'Amplitude'})
#    model.append({'element':'variable','name':'offset','write':self.offset,'read':self.get_offset,'unit':'V','type':float,'help':'Offset'})
#    model.append({'element':'variable','name':'frequency','write':self.frequency,'read':self.get_frequency,'unit':'Hz','type':float,'help':'Frequency'})

#    return model



# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 14:54:41 2024

@author: Jonathan
"""

import re
from typing import Any, List, Tuple

import numpy as np
import pandas as pd

from .devices import DEVICES
from .utilities import clean_string


# class AddVarSignal(QtCore.QObject):
#     add = QtCore.Signal(object, object)
#     def emit_add(self, name, value):
#         self.add.emit(name, value)


# class RemoveVarSignal(QtCore.QObject):
#     remove = QtCore.Signal(object)
#     def emit_remove(self, name):
#         self.remove.emit(name)


# class MyDict(dict):

#     def __init__(self):
#         self.addVarSignal = AddVarSignal()
#         self.removeVarSignal = RemoveVarSignal()

#     def __setitem__(self, item, value):
#         super(MyDict, self).__setitem__(item, value)
#         self.addVarSignal.emit_add(item, value)

#     def pop(self, item):
#         super(MyDict, self).pop(item)
#         self.removeVarSignal.emit_remove(item)


# VARIABLES = MyDict()
VARIABLES = {}

EVAL = "$eval:"


def update_allowed_dict() -> dict:
    global allowed_dict  # needed to remove variables instead of just adding new one
    allowed_dict = {"np": np, "pd": pd}
    allowed_dict.update(DEVICES)
    allowed_dict.update(VARIABLES)
    return allowed_dict


allowed_dict = update_allowed_dict()

# OPTIMIZE: Variable becomes closer and closer to core.elements.Variable, could envision a merge
# TODO: refresh menu display by looking if has eval (no -> can refresh)
# TODO add read signal to update gui (separate class for event and use it on itemwidget creation to change setText with new value)
class Variable():
    """ Class used to control basic variable """

    raw: Any
    value: Any

    def __init__(self, name: str, var: Any):
        """ name: name of the variable, var: value of the variable """
        self.unit = None
        self.writable = True
        self.readable = True
        self._rename(name)
        self.write_function(var)

    def _rename(self, new_name: str):
            self.name = new_name
            self.address = lambda: new_name

    def write_function(self, var: Any):
        if isinstance(var, Variable):
            self.raw = var.raw
            self.value = var.value
        else:
            self.raw = var
            self.value = 'Need update' if has_eval(self.raw) else self.raw

        # If no devices or variables with char '(' found in raw, can evaluate value safely
        if not has_variable(self.raw) or '(' not in self.raw:
            try: self.value = self.read_function()
            except Exception as e: self.value = str(e)

        self.type = type(self.raw)  # For slider

    def read_function(self):
        if has_eval(self.raw):
            value = str(self.raw)[len(EVAL): ]
            call = eval(str(value), {}, allowed_dict)
            self.value = call
        else:
            call = self.value

        return call

    def __call__(self, value: Any = None) -> Any:
        if value is None:
            return self.read_function()

        self.write_function(value)
        return None


def list_variables() -> List[str]:
    ''' Returns a list of Variables '''
    return list(VARIABLES)


def rename_variable(name: str, new_name: str):
    ''' Rename an existing Variable '''
    new_name = clean_string(new_name)
    var = VARIABLES.pop(name)
    VARIABLES[new_name] = var
    var._rename(new_name)
    update_allowed_dict()


def set_variable(name: str, value: Any) -> Variable:
    ''' Create or modify a Variable with provided name and value '''
    name = clean_string(name)

    if is_Variable(value):
        var = value
        var(value)
    else:
        if name in VARIABLES:
            var = get_variable(name)
            var(value)
        else:
            var = Variable(name, value)

    VARIABLES[name] = var
    update_allowed_dict()
    return var


def get_variable(name: str) -> Variable:
    ''' Return Variable with provided name '''
    assert name in VARIABLES, f"Variable name '{name}' not found in {list_variables()}"
    return VARIABLES[name]


def remove_variable(name: str) -> Variable:
    var = VARIABLES.pop(name)
    update_allowed_dict()
    return var


def remove_from_config(variables: List[Tuple[str, Any]]):
    for name, _ in variables:
        if name in VARIABLES:
            remove_variable(name)


def update_from_config(variables: List[Tuple[str, Any]]):
    for var in variables:
        set_variable(var[0], var[1])


def has_variable(value: str) -> bool:
    if not isinstance(value, str): return False
    if has_eval(value): value = value[len(EVAL): ]

    pattern = r'[a-zA-Z_][a-zA-Z0-9_]*(?:\.[a-zA-Z_][a-zA-Z0-9_]*)*'
    pattern_match = [var.split('.')[0] for var in re.findall(pattern, value)]

    for key in (list(DEVICES) + list(VARIABLES)):
        if key in pattern_match:
            return True
    return False


def has_eval(value: Any) -> bool:
    """ Checks if value is a string starting with '$eval:'"""
    return True if isinstance(value, str) and value.startswith(EVAL) else False


def is_Variable(value: Any):
    """ Returns True if value of type Variable """
    return isinstance(value, Variable)


def eval_variable(value: Any) -> Any:
    """ Evaluate the given python string. String can contain variables,
    devices, numpy arrays and pandas dataframes."""
    if has_eval(value): value = Variable('temp', value)

    if is_Variable(value): return value()
    return value


def eval_safely(value: Any) -> Any:
    """ Same as eval_variable but do not evaluate if contains devices or variables """
    if has_eval(value): value = Variable('temp', value)

    if is_Variable(value): return value.value
    return value



# -*- coding: utf-8 -*-
"""
Created on Tue Nov 21 18:16:20 2023

@author: Jonathan
"""

import os
import sys
import zipfile
import tempfile
import shutil
import urllib.request
import json
from typing import Union, Tuple

from .paths import DRIVER_SOURCES, DRIVER_REPOSITORY
from .drivers import update_drivers_paths
from .utilities import input_wrap
from .gitdir import download


def _format_url(url: str):
    """ Change github repo name to download link """

    if url.endswith(".zip"):
        format_url = url
    else:
        format_url = url if "/tree/" in url else url + "/tree/master"
        format_url = format_url.replace("/tree/", "/archive/refs/heads/")
        format_url += ".zip"
    return format_url


def _download_repo(url: str, output_dir: str):
    try:
        import requests
        from tqdm import tqdm
    except:
        print("Package tqdm or requests not found, can't display download progression")
        print(f"Downloading {url}")
        with urllib.request.urlopen(url) as github_repo_zip:
            with open(output_dir, 'wb') as repo_zip:
                repo_zip.write(github_repo_zip.read())
    else:
        """https://stackoverflow.com/questions/37573483/progress-bar-while-download-file-over-http-with-requests"""
        # Streaming, so we can iterate over the response.
        response = requests.get(url, stream=True)

        # Sizes in bytes.
        total_size = int(response.headers.get("content-length", 0))
        block_size = 1024

        with tqdm(total=total_size, unit="B", unit_scale=True,
                  desc=f"Downloading {url}") as progress_bar:
            with open(output_dir, "wb") as file:
                for data in response.iter_content(block_size):
                    progress_bar.update(len(data))
                    file.write(data)

        if total_size not in (0, progress_bar.n):
            raise RuntimeError("Could not download file")


def _unzip_repo(repo_zip: str, output_dir: str):
    """ Unzip repo_zip to output_dir using a temporary folder"""
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)

    temp_dir = tempfile.mkdtemp()
    try:
        from tqdm import tqdm
    except:
        print("Package tqdm not found, can't display progression")
        with zipfile.ZipFile(repo_zip, 'r') as zip_open:
            repo_name = zip_open.namelist()[0][:-1]
            print(f'Extracting {repo_name}')
            zip_open.extractall(temp_dir)

        temp_unzip_repo = os.path.join(temp_dir, repo_name)
        print(f'Moving drivers to {output_dir}')
        for filename in os.listdir(temp_unzip_repo):
            _copy_move(temp_unzip_repo, filename, output_dir)
    else:
        """https://stackoverflow.com/questions/4341584/extract-zipfile-using-python-display-progress-percentage"""
        with zipfile.ZipFile(repo_zip, 'r') as zf:
            repo_name = str(zf.namelist()[0])[:-1]
            for member in tqdm(zf.infolist(),
                               desc=f'Extracting {repo_name}'):
                zf.extract(member, temp_dir)

        temp_unzip_repo = os.path.join(temp_dir, repo_name)
        for filename in tqdm(os.listdir(temp_unzip_repo),
                             desc=f'Moving drivers to {output_dir}'):
            _copy_move(temp_unzip_repo, filename, output_dir)

    os.rmdir(temp_unzip_repo)
    os.rmdir(temp_dir)


def _copy_move(temp_unzip_repo, filename, output_dir):
    temp_file = os.path.join(temp_unzip_repo, filename)
    output_file = os.path.join(output_dir, filename)

    if os.path.isfile(temp_file):
        shutil.copy(temp_file, output_file)
        os.remove(temp_file)
    else:
        try:
            shutil.copytree(temp_file, output_file, dirs_exist_ok=True)  # python >=3.8 only
        except:
            if os.path.exists(output_file):
                shutil.rmtree(output_file, ignore_errors=True)
            shutil.copytree(temp_file, output_file)

        shutil.rmtree(temp_file, ignore_errors=True)


def _check_empty_driver_folder():
    if not os.listdir(DRIVER_SOURCES['official']):
        print(f"No drivers found in {DRIVER_SOURCES['official']}")
        install_drivers()


def install_drivers(*repo_url: Union[None, str, Tuple[str, str]],
                    skip_input=False):
    """ Ask if want to install drivers from repo url.
    repo_url: can be url or tuple ('path to install', 'url to download').
    If no argument passed, download official drivers to official driver folder.
    If only url given, use official driver folder.
    Also install mandatory drivers (system, dummy, plotter) from official repo."""
    # Download mandatory drivers
    official_folder = DRIVER_SOURCES['official']
    official_url = DRIVER_REPOSITORY[official_folder]
    mandatory_drivers = ['system', 'dummy', 'plotter']

    for driver in mandatory_drivers:
        if not os.path.isdir(os.path.join(official_folder, driver)):
            print(f"Installing driver '{driver}' to {official_folder}")
            _download_driver(official_url, driver, official_folder, _print=False)

    # Ask if want to download drivers from repo_url
    temp_folder = os.environ['TEMP']  # This variable can be set in autolab_config.ini
    temp_repo_folder = tempfile.mkdtemp(dir=temp_folder)

    # create list of tuple with tuple being ('path to install', 'url to download')
    if len(repo_url) == 0:
        list_repo_tuple = list(DRIVER_REPOSITORY.items())  # This variable can be modified in autolab_config.ini
    else:
        list_repo_tuple = list(repo_url)
        for i, repo_url_tmp in enumerate(list_repo_tuple):
            if isinstance(repo_url_tmp, str):
                list_repo_tuple[i] = (official_folder, repo_url_tmp)
            elif isinstance(repo_url_tmp, dict):
                raise TypeError("Error: This option has been removed, use tuple instead with (folder, url)")
            elif not isinstance(repo_url_tmp, tuple):
                raise TypeError(f'repo_url must be str or tuple. Given {type(repo_url_tmp)}')
            assert len(list_repo_tuple[i]) == 2, "Expect (folder, url), got wrong length: {len(list_repo_tuple[i])} for {list_repo_tuple[i]}"

    for repo_tuple in list_repo_tuple:
        drivers_folder, drivers_url = repo_tuple

        if r"github.com/" in drivers_url:
            repo_name = drivers_url.split(r"github.com/")[1].split("/")[1]
        else:
            repo_name = "temp_folder"

        repo_zip_url = _format_url(drivers_url)
        zip_name = "-".join((repo_name, os.path.basename(repo_zip_url)))
        temp_repo_zip = os.path.join(temp_repo_folder, zip_name)

        if skip_input: ans = 'yes'
        else:
            ans = input_wrap(f'Install drivers from {drivers_url} to {drivers_folder}? [default:yes] > ')

        if ans.strip().lower() == 'no': continue

        _download_repo(repo_zip_url, temp_repo_zip)
        _unzip_repo(temp_repo_zip, drivers_folder)
        os.remove(temp_repo_zip)
    os.rmdir(temp_repo_folder)

    # Update available drivers
    update_drivers_paths()


# =============================================================================
# Functions used to download specific driver, Not optimized to download full
# repository. Use install_drivers() instead
# =============================================================================

def _get_drivers_list_from_github(url):
    """ Returns driver list from a github repo.
    url of a repository contaning driver folders.
    """
    url_api = url

    if url_api.startswith('https://github.com'):
        url_api = url_api.replace('https://github.com', 'https://api.github.com/repos')

    if '/tree/' in url_api:
        url_api = url_api.replace("/tree/", '/git/trees/')
    else:
        url_api += '/git/trees/master'

    with urllib.request.urlopen(url_api) as f:
        html = json.load(f)

    driver_list = [tree['path'] for tree in html['tree'] if '.' not in tree['path']]
    if 'LICENSE' in driver_list:
        driver_list.remove('LICENSE')

    return driver_list

def _download_driver(url, driver_name, output_dir, _print=True):
    """ Download a driver from an github url """
    try:
        if _print:
            print(f"Downloading {driver_name}")

        driver_url = url
        if '/tree/' not in driver_url: driver_url += '/tree/master'
        driver_url = driver_url + "/" + driver_name
        # Too slow to be used for full repo, only use it for one or 2 drivers
        # 'HTTP Error 403: rate limit exceeded' due to too much download if don't have github account
        download(driver_url, output_dir=output_dir, _print=_print)
    except:  # if use Exception, crash python when having error
        e = f"Error when downloading driver '{driver_name}'"
        if _print:
            print(e, file=sys.stderr)
        else:
            return e



# -*- coding: utf-8 -*-
import sys

from .config import get_device_config
from .drivers import (update_drivers_paths, DRIVERS_PATHS, get_driver_category,
                      load_driver_lib, get_connection_names, get_class_args,
                      get_connection_class, get_driver_class, get_module_names,
                      get_module_class)
from .paths import DRIVER_SOURCES
from .utilities import two_columns, emphasize, underline
from .devices import list_devices, list_loaded_devices, get_final_device_config

# =============================================================================
# INFOS
# =============================================================================

def _list_drivers(_print: bool = True) -> str:
    ''' Returns a list of all the drivers with categories by sections
    (autolab drivers, local drivers) '''
    update_drivers_paths()

    s = '\n'
    s += f'{len(DRIVERS_PATHS)} drivers found\n\n'

    for i, (source_name, source) in enumerate(DRIVER_SOURCES.items()):
        sub_driver_list = sorted([key for key, val in DRIVERS_PATHS.items(
            ) if val['source'] == source_name])
        s += f'Drivers in {source}:\n'
        if len(sub_driver_list) > 0:
            txt_list = [[f' - {driver_name}',
                         f'({get_driver_category(driver_name)})']
                            for driver_name in sub_driver_list]
            s += two_columns(txt_list) + '\n\n'
        else:
            if (i + 1) == len(DRIVER_SOURCES):
                s += ' <No drivers>\n\n'
            else:
                s += ' <No drivers> (or overwritten)\n\n'

    if _print:
        print(s)
        return None
    return s


def _list_devices(_print: bool = True) -> str:
    ''' Returns a list of all the devices and their associated drivers
    from devices_config.ini '''
    # Gather local config informations
    devices_names = list_devices()
    devices_names_loaded = list_loaded_devices()

    # Build infos str for devices
    s = '\n'
    s += f'{len(devices_names)} devices found\n\n'
    txt_list = [
        [f' - {name} ' + ('[loaded]' if name in devices_names_loaded else ''),
         f'({get_device_config(name)["driver"]})'] for name in devices_names]
    s += two_columns(txt_list) + '\n'

    if _print:
        print(s)
        return None
    return s


def infos(_print: bool = True) -> str:
    ''' Returns a list of all the drivers and all the devices,
    along with their associated drivers from devices_config.ini '''
    s  = ''
    s += _list_drivers(_print=False)
    s += _list_devices(_print=False)

    if _print:
        print(s)
        return None
    return s

# =============================================================================
# DRIVERS
# =============================================================================

def config_help(driver_name: str, _print: bool = True, _parser: bool = False) -> str:
    ''' Display the help of a particular driver (connection types, modules, ...) '''
    try:
        driver_name = get_final_device_config(driver_name)["driver"]
    except:
        pass
    # Load list of all parameters
    try:
        driver_lib = load_driver_lib(driver_name)
    except Exception as e:
        print(f"Can't load {driver_name}: {e}", file=sys.stderr)
        return None
    params = {}
    params['driver'] = driver_name
    params['connection'] = {}
    for conn in get_connection_names(driver_lib):
        params['connection'][conn] = get_class_args(
            get_connection_class(driver_lib, conn))
    params['other'] = get_class_args(get_driver_class(driver_lib))
    if hasattr(get_driver_class(driver_lib), 'slot_config'):
        params['other']['slot1'] = f'{get_driver_class(driver_lib).slot_config}'
        params['other']['slot1_name'] = 'my_<MODULE_NAME>'

    mess = '\n'

    # Name and category if available
    submess = f'Driver "{driver_name}" ({get_driver_category(driver_name)})'
    mess += emphasize(submess, sign='=') + '\n'

    # Connections types
    c_option=' (-C option)' if _parser else ''
    mess += f'\nAvailable connections types{c_option}:\n'
    for connection in params['connection']:
        mess += f' - {connection}\n'
    mess += '\n'

    # Modules
    if hasattr(get_driver_class(driver_lib), 'slot_config'):
        mess += 'Available modules:\n'
        modules = get_module_names(driver_lib)
        for module in modules:
            moduleClass = get_module_class(driver_lib, module)
            mess += f' - {module}'
            if hasattr(moduleClass, 'category'): mess += f' ({moduleClass.category})'
            mess += '\n'
        mess += '\n'

    # Example of a devices_config.ini section
    mess += '\n' + underline(
        'Saving a Device configuration in devices_config.ini:') + '\n'
    for conn in params['connection']:
        mess += f"\n[my_{params['driver']}]\n"
        mess += f"driver = {params['driver']}\n"
        mess += f"connection = {conn}\n"
        for arg, value in params['connection'][conn].items():
            mess += f"{arg} = {value}\n"
        for arg, value in params['other'].items():
            mess += f"{arg} = {value}\n"

    # Example of get_driver
    mess += '\n\n' + underline('Loading a Driver:') + '\n\n'
    for conn in params['connection']:
        if not _parser:
            args_str = f"'{params['driver']}', connection='{conn}'"
            for arg, value in params['connection'][conn].items():
                args_str += f", {arg}='{value}'"
            for arg, value in params['other'].items():
                if isinstance(value, str):
                    args_str += f", {arg}='{value}'"
                else:
                    args_str += f", {arg}={value}"
            mess += f"a = autolab.get_driver({args_str})\n"
        else:
            args_str = f"-D {params['driver']} -C {conn} "
            for arg,value in params['connection'][conn].items():
                if arg == 'address': args_str += f"-A {value} "
                if arg == 'port': args_str += f"-P {value} "
            if len(params['other']) > 0: args_str += '-O '
            for arg,value in params['other'].items():
                args_str += f"{arg}={value} "
            mess += f"autolab driver {args_str} -m method(value) \n"

    # Example of get_device
    mess += '\n\n' + underline(
        'Loading a Device configured in devices_config.ini:') + '\n\n'
    if not _parser:
        mess += f"a = autolab.get_device('my_{params['driver']}')"
    else:
        mess += f"autolab device -D my_{params['driver']} -e element -v value \n"

    if _print:
        print(mess)
        return None
    return mess



# -*- coding: utf-8 -*-
"""
Created on Mon Nov 18 14:53:10 2019

@author: qchat
"""

import os
import tempfile
import configparser
from typing import List

from .paths import PATHS, DRIVER_SOURCES, DRIVER_REPOSITORY
from .utilities import boolean


# =============================================================================
# GENERAL
# =============================================================================

def initialize_local_directory() -> bool:
    """ This function creates the default autolab local directory.
    Returns True if create default autolab folder (first autolab use) """
    FIRST = False
    _print = True
    # LOCAL DIRECTORY
    if not os.path.exists(PATHS['user_folder']):
        os.mkdir(PATHS['user_folder'])
        print(f"The local directory of AUTOLAB has been created: {PATHS['user_folder']}.\n"\
              'It contains the configuration files devices_config.ini, autolab_config.ini ' \
              'and plotter.ini.\n' \
              "It also contains the 'driver' directory with 'official' and 'local' sub-directories."
              )
        _print = False
        FIRST = True

    # DEVICES CONFIGURATION FILE
    if not os.path.exists(PATHS['devices_config']):
        devices_config = configparser.ConfigParser()
        devices_config['system'] = {'driver': 'system', 'connection': 'DEFAULT'}
        devices_config['dummy'] = {'driver': 'dummy', 'connection': 'CONN'}
        devices_config['plotter'] = {'driver': 'plotter', 'connection': 'DEFAULT'}
        save_config('devices_config', devices_config)
        if _print: print(f"The devices configuration file devices_config.ini has been created: {PATHS['devices_config']}")

    # DRIVER FOLDERS
    if not os.path.exists(PATHS['drivers']):
        os.mkdir(PATHS['drivers'])
        if _print: print(f"The drivers directory has been created: {PATHS['drivers']}")
    if not os.path.exists(DRIVER_SOURCES['official']):
        os.mkdir(DRIVER_SOURCES['official'])
        if _print: print(f'The official driver directory has been created: {DRIVER_SOURCES["official"]}')
    if not os.path.exists(DRIVER_SOURCES['local']):
        os.mkdir(DRIVER_SOURCES['local'])
        if _print: print(f'The local driver directory has been created: {DRIVER_SOURCES["local"]}')

    # AUTOLAB CONFIGURATION FILE
    if not os.path.exists(PATHS['autolab_config']):
        save_config('autolab_config', configparser.ConfigParser())
        if _print: print(f"The configuration file autolab_config.ini has been created: {PATHS['autolab_config']}")

    # PLOTTER CONFIGURATION FILE
    if not os.path.exists(PATHS['plotter_config']):
        save_config('plotter_config', configparser.ConfigParser())
        if _print: print(f"The configuration file plotter_config.ini has been created: {PATHS['plotter_config']}")

    return FIRST


def save_config(config_name: str, config: configparser.ConfigParser):
    """ This function saves the given config parser in the autolab configuration file """
    with open(PATHS[config_name], 'w') as file:
        config.write(file)


def load_config(config_name: str) -> configparser.ConfigParser:
    """ This function loads the autolab configuration file in a config parser """
    config = configparser.ConfigParser(allow_no_value=True, delimiters='=')  # don't want ':' as delim, needed for path as key
    config.optionxform = str
    try:  # encoding order matter
        config.read(PATHS[config_name],
                    encoding='utf-8')
    except:
        config.read(PATHS[config_name])

    return config


def modify_config(config_name: str, config_dict: dict) -> configparser.ConfigParser:
    """ Returns a modified config file structures using the input dict """
    config = load_config(config_name)

    for section_key, section_dic in config_dict.items():
        conf = {}
        for key, dic in section_dic.items():
            conf[key] = str(dic)

        config[section_key] = conf

    return config


# def get_attribute(config,section_name,attribute_name):

#     ''' This function get the value of the given parameter of the given header in the provided configparser '''
#     config_section = get_config_section(config,section_name)
#     assert attribute_name in config_section.keys(), f"Attribute '{attribute_name}' not found in section {section_name}."
#     return config_section[attribute_name]

# def get_config_section(config,section_name):

#     ''' Returns section <section_name> from existing <config> object '''

#     assert section_name in config.sections(), f"Section {section_name} not found in configuration file"
#     return config[section_name]


# def load_config_section(config_name,section_name):

#     ''' Returns section <section_name> from configuration file <config_name>_config.ini '''

#     config = load_config(config_name)
#     return get_config_section(config,section_name)


# =============================================================================
# AUTOLAB CONFIG
# =============================================================================

autolab_dict = {
    'server': {'port': 4001},
    'GUI': {'qt_api': "default",
            'theme': "default",
            'font_size': 10,
            'image_background': 'w',
            'image_foreground': 'k',
            },
    'control_center': {'precision': 7,
                       'print': True,
                       'logger': False,
                       'console': False,
                       },
    'monitor': {'precision': 4,
                'save_figure': True},
    'scanner': {'precision': 15,
                'save_config': True,
                'save_figure': True,
                'save_temp': True,
                'ask_close': True,
                },
    'directories': {'temp_folder': 'default'},
    'extra_driver_path': {},
    'extra_driver_url_repo': {},
}


def change_autolab_config(config: configparser.ConfigParser):
    """ Save the autolab config file structures with comments """
    config.set('GUI', '# qt_api -> Choose between default, pyqt5, pyside2, pyqt6 and pyside6')
    config.set('GUI', '# theme -> Choose between default and dark')
    config.set('scanner', '# Think twice before using save_temp = False')
    config.set('extra_driver_path', r'# Example: onedrive = C:\Users\username\OneDrive\my_drivers')
    config.set('extra_driver_url_repo', r'# Example: C:\Users\username\OneDrive\my_drivers = https://github.com/my_repo/my_drivers')

    if not boolean(config['scanner']["save_temp"]):
        print('Warning: save_temp in "autolab_config.ini" is disabled, ' \
              'be aware that data will not be saved during the scan. ' \
              'If a crash occurs during a scan, you will loose its data. ' \
              'Disabling this option is only useful if you want to do fast ' \
              'scan when plotting large dataframe (images for examples)')

    save_config('autolab_config', config)


def check_autolab_config():
    """ Changes the autolab config file structures """
    autolab_config = load_config('autolab_config')

    for section_key, section_dic in autolab_dict.items():
        if section_key in autolab_config.sections():
            conf = dict(autolab_config[section_key])
            for key, value in section_dic.items():
                if key not in conf:
                    conf[key] = str(value)
        else:
            conf = section_dic

        autolab_config[section_key] = conf

    # added in 2.0 for retrocompatibilty with 1.1.12
    if 'QT_API' in autolab_config['GUI']:
        value = autolab_config.get('GUI', 'QT_API')
        autolab_config.remove_option('GUI', 'QT_API')
        autolab_config.set('GUI', 'qt_api', value)

    # Check and correct boolean, float and int
    for section_key, section_dic in autolab_dict.items():
        for key, value in section_dic.items():
            try:
                if isinstance(value, bool):
                    boolean(autolab_config[section_key][key])
                elif isinstance(value, float):
                    float(autolab_config[section_key][key])
                elif isinstance(value, int):
                    int(float(autolab_config[section_key][key]))
            except:
                autolab_config[section_key][key] = str(autolab_dict[section_key][key])
                print(f'Wrong {section_key} {key} in config, change to default value')

    # Check for specific values
    if autolab_config['GUI']['theme'] not in ('default', 'dark'):
        autolab_config['GUI']['theme'] = str(autolab_dict['GUI']['theme'])
        print('Wrong GUI theme in config, change to default value')

    change_autolab_config(autolab_config)


def get_config(section_name: str) -> configparser.SectionProxy:
    ''' Returns section from autolab_config.ini '''
    config = load_config('autolab_config')
    assert section_name in config.sections(), f'Missing {section_name} section in autolab_config.ini'
    return config[section_name]


def get_server_config() -> configparser.SectionProxy:
    ''' Returns section server from autolab_config.ini '''
    return get_config('server')


def get_GUI_config() -> configparser.SectionProxy:
    ''' Returns section qt_api from autolab_config.ini '''
    return get_config('GUI')


def get_control_center_config() -> configparser.SectionProxy:
    ''' Returns section control_center from autolab_config.ini '''
    return get_config('control_center')


def get_monitor_config() -> configparser.SectionProxy:
    ''' Returns section monitor from autolab_config.ini '''
    return get_config('monitor')


def get_scanner_config() -> configparser.SectionProxy:
    ''' Returns section scanner from autolab_config.ini '''
    return get_config('scanner')


def get_directories_config() -> configparser.SectionProxy:
    ''' Returns section directories from autolab_config.ini '''
    return get_config('directories')


def get_extra_driver_path_config() -> configparser.SectionProxy:
    ''' Returns section extra_driver_path from autolab_config.ini '''
    return get_config('extra_driver_path')


def get_extra_driver_repo_url_config() -> configparser.SectionProxy:
    ''' Returns section extra_driver_url_repo from autolab_config.ini '''
    return get_config('extra_driver_url_repo')


def set_temp_folder() -> str:
    ''' Set temporary folder using given path in autolab_config.ini
    Write it in os.environ['TEMP'] to be used in autolab and drivers '''
    temp_folder = get_directories_config()["temp_folder"]

    if temp_folder == 'default':
        # Try to get TEMP, if not get tempfile
        temp_folder = os.environ.get('TEMP', tempfile.gettempdir())

    # Always write temp to allow drivers to get and change it
    os.environ['TEMP'] = temp_folder

    return temp_folder


def add_extra_driver_path():
    """ Add extra driver path to DRIVER_SOURCES. If several drivers have the
    same name among different folders, the last one called will be used """
    extra_driver_path = get_extra_driver_path_config()

    for driver_path_name in extra_driver_path.keys():
        assert driver_path_name not in ['official', 'local'], (
            "Can't change 'official' nor 'local' driver folder paths. " \
            f"Change name '{driver_path_name}' in section [extra_driver_path] of {PATHS['autolab_config']}")

    DRIVER_SOURCES.update(extra_driver_path)


def add_extra_driver_repo_url():
    """ Add extra driver repository url to DRIVER_REPOSITORY.If several drivers
    have the same name, the last one overwrites the others (destructive operation) """
    extra_driver_path = get_extra_driver_repo_url_config()

    for driver_path_name in extra_driver_path.keys():
        assert driver_path_name not in [DRIVER_SOURCES['official']], (
            "Can't install driver in 'official' folder. " \
            f"Change path '{driver_path_name}' in section [extra_driver_path]" \
            f" of {PATHS['autolab_config']}")

        DRIVER_REPOSITORY.update(
            {driver_path_name: extra_driver_path[driver_path_name]})


# =============================================================================
# PLOTTER CONFIG
# =============================================================================

plotter_dict = {
    'plugin': {'plotter': 'plotter'},
    'device': {'address': 'dummy.array_1D'},
}

def change_plotter_config(config: configparser.ConfigParser):
    """ Save the plotter config file structures with comments """
    config.set('plugin', '# Usage: <PLUGIN_NAME> = <DEVICE_NAME>')
    config.set('plugin', '# Example: plotter = plotter')
    config.set('device', '# Usage: address = <DEVICE_VARIABLE>')
    config.set('device', '# Example: address = dummy.array_1D')

    save_config('plotter_config', config)


def check_plotter_config():
    """ This function checks config file structures """
    plotter_config = load_config('plotter_config')

    for section_key, section_dic in plotter_dict.items():
        if section_key in plotter_config.sections():
            conf = dict(plotter_config[section_key])
            for key, value in section_dic.items():
                if key not in conf:
                    conf[key] = str(value)
        else:
            conf = section_dic

        plotter_config[section_key] = conf

    change_plotter_config(plotter_config)


# =============================================================================
# DEVICES CONFIG
# =============================================================================

def get_all_devices_configs() -> configparser.ConfigParser:
    ''' Returns current devices configuration '''
    config = load_config('devices_config')
    assert len(set(config.sections())) == len(config.sections()), "Each device must have a unique name."
    return config


def list_all_devices_configs() -> List[str]:
    ''' Returns the list of available configuration names '''
    devices_configs = get_all_devices_configs()
    return sorted(list(devices_configs.sections()))


def get_device_config(config_name) -> configparser.SectionProxy:
    ''' Returns the config associated with config_name '''
    assert config_name in list_all_devices_configs(), f"Device configuration {config_name} not found"
    devices_configs = get_all_devices_configs()
    return devices_configs[config_name]
